from pathlib import Path
from typing import cast

import numpy as np

from rock_physics_open.equinor_utilities import gen_utilities
from rock_physics_open.equinor_utilities.optimisation_utilities import (
    gen_mod_routine,
    gen_sub_routine,
    load_opt_params,
    opt_param_info,
)
from rock_physics_open.equinor_utilities.various_utilities.types import Array1D

from .curvefit_t_matrix_exp import curvefit_t_matrix_exp


def run_t_matrix_with_opt_params_exp(
    fl_k_orig: Array1D[np.float64],
    fl_rho_orig: Array1D[np.float64],
    fl_k_sub: Array1D[np.float64],
    fl_rho_sub: Array1D[np.float64],
    vp: Array1D[np.float64],
    vs: Array1D[np.float64],
    rhob: Array1D[np.float64],
    phi: Array1D[np.float64],
    vsh: Array1D[np.float64],
    angle: float,
    perm: float,
    visco: float,
    tau: float,
    freq: float,
    f_name: str | Path,
    fluid_sub: bool = True,
) -> tuple[
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
]:
    """Based on the input file with parameters for the optimally fitted model, the correct modelling version is run.
    Fluid substitution follows, in case it is selected. If not, the vp_sub and vs_sub will contain the same values as
    the input logs.

    Parameters
    ----------
    fl_k_orig : np.ndarray
        Effective in situ fluid bulk modulus [Pa].
    fl_rho_orig : np.ndarray
        Effective in situ fluid density [kg/m^3].
    fl_k_sub : np.ndarray
        Effective substituted fluid bulk modulus [Pa].
    fl_rho_sub : np.ndarray
        Effective substituted density [kg/m^3].
    vp : np.ndarray
        Compressional velocity [m/s].
    vs : np.ndarray
        Shear velocity [m/s].
    rhob : np.ndarray
        Bulk density [kg/m^3].
    phi : np.ndarray
        Porosity [fraction].
    vsh : np.ndarray
        Shale volume [fraction].
    angle : float
        Angle of symmetry plane [degrees]
    perm : float
        Permeability [mD].
    visco : float
        Viscosity [cP].
    tau : float
        Relaxation time constant [s].
    freq : float
        Signal frequency [Hz].
    f_name : str
        File name for parameter file for optimal parameters.
    fluid_sub : bool
        Boolean parameter to perform fluid substitution.

    Returns
    -------
    tuple
        Tuple of np.ndarrays: vp and vs for pressure substituted case, vp, vs and density for fluid substituted case, vp and vs for
        optimal fitted model, vp and vs residuals (observed logs minus modelled values).
    """

    opt_type, opt_params, opt_dict = load_opt_params(f_name)
    y_data = np.stack([vp, vs], axis=1)
    y_shape = y_data.shape
    phi, angle_, perm_, visco_, tau_, freq_, def_vpvs = cast(
        list[Array1D[np.float64]],
        gen_utilities.dim_check_vector((phi, angle, perm, visco, tau, freq, 1.0)),
    )

    rho_sub = rhob + (fl_rho_sub - fl_rho_orig) * phi
    # Set None values for inputs that will be defined in the different cases
    x_data_new = None

    opt_fcn = curvefit_t_matrix_exp
    _, scale_val, _ = opt_param_info()
    # Generate x_data according to method exp
    x_data = np.stack(
        (
            phi,
            vsh,
            fl_k_orig,
            fl_rho_orig,
            angle_,
            perm_,
            visco_,
            tau_,
            freq_,
            def_vpvs,
        ),
        axis=1,
    )
    if fluid_sub:
        x_data_new = np.stack(
            (
                phi,
                vsh,
                fl_k_sub,
                fl_rho_sub,
                angle_,
                perm_,
                visco_,
                tau_,
                freq_,
                def_vpvs,
            ),
            axis=1,
        )

        v_sub, v_mod, v_res = gen_sub_routine(
            opt_function=opt_fcn,
            xdata_orig=x_data,
            xdata_new=x_data_new,
            ydata=y_data,
            opt_params=opt_params,
        )
        vp_sub, vs_sub = [arr.flatten() for arr in np.split(v_sub, 2, axis=1)]
        vp_mod, vs_mod = [arr.flatten() for arr in np.split(v_mod, 2, axis=1)]
        vp_res, vs_res = [arr.flatten() for arr in np.split(v_res, 2, axis=1)]
    else:
        v_mod = gen_mod_routine(opt_fcn, x_data, y_shape, opt_params)
        vp_mod, vs_mod = [arr.flatten() for arr in np.split(v_mod, 2, axis=1)]
        vp_sub = vp
        vs_sub = vs
        vp_res = vp_mod - vp
        vs_res = vs_mod - vs

    if opt_type != "exp":
        raise TypeError(
            f"{__file__}: incorrect type of optimal parameter input file, must come from EXP optimisation"
        )

    rho_mod = (
        (1.0 - vsh) * opt_dict["rho_carb"] * scale_val["rho_carb"]  # pyright: ignore[reportTypedDictNotRequiredAccess] | Should be there for exp
        + vsh * opt_dict["rho_sh"] * scale_val["rho_sh"]  # pyright: ignore[reportTypedDictNotRequiredAccess] | Should be there for exp
    ) * (1.0 - phi) + phi * fl_rho_orig

    rho_res = rho_mod - rhob
    ai_sub = vp_sub * rho_sub
    vpvs_sub = vp_sub / vs_sub

    return (
        vp_sub,
        vs_sub,
        rho_sub,
        ai_sub,
        vpvs_sub,
        vp_mod,
        vs_mod,
        rho_mod,
        vp_res,
        vs_res,
        rho_res,
    )



import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array2D,
    Array3D,
    Array4D,
)

from .array_functions import array_inverse, array_matrix_mult
from .g_tensor import g_tensor_vec


def calc_kd_vec(
    c0: Array3D[np.float64],
    i4: Array3D[np.float64],
    s_0: Array3D[np.float64],
    alpha: Array2D[np.float64],
) -> Array4D[np.float64]:
    """
    kd is a (nx6x6x(number of alphas)) matrix.

    Parameters
    ----------
    c0 : np.ndarray
        Stiffness tensor of the host material (nx6x6 matrix).
    i4 : np.ndarray
        Array of 6x6 identity matrices.
    s_0:  np.ndarray
        Inverse of stiffness tensor.
    alpha : np.ndarray
        Vector of aspect ratios (1x (number of aspect ratios) vector) or nx(number of (number of alphas)).

    Returns
    -------
    np.ndarray
        kd: stiffness tensor.
    """
    log_length = c0.shape[0]

    if alpha.ndim == 1 and alpha.shape[0] != c0.shape[0]:
        alpha = np.tile(alpha.reshape(1, alpha.shape[0]), (log_length, 1))
    L = alpha.shape[1]
    kd = np.zeros((log_length, 6, 6, L))

    for nc in range(L):
        g = g_tensor_vec(
            c0=c0,
            s_0=s_0,
            alpha=alpha[:, nc],
        )
        kd[:, :, :, nc] = array_matrix_mult(
            array_inverse(i4 + array_matrix_mult(g, c0)), s_0
        )

    return kd



import numpy as np

from rock_physics_open.equinor_utilities.std_functions import (
    gassmann,
    gassmann2,
    gassmann_dry,
)


class TestGassmann:
    def setup_gassmann(self):
        k_brine = np.array([2.8e9, 2.8e9, 2.8e9])
        phie = np.array([0.0, 0.35, 0.1])
        k_min = np.array([36.6e9, 36.6e9, 36.6e9])
        k_dry = np.array([36.6e9, 15.0e9, 28.0e9])
        k_oil = np.array([0.8e9, 0.8e9, 0.8e9])
        return k_brine, phie, k_min, k_dry, k_oil

    def test_gassmann(self):
        k_brine, phie, k_min, k_dry, _ = self.setup_gassmann()
        k_sat = gassmann(k_dry, phie, k_brine, k_min)
        ksat_ref = np.array([36.6, 17.6473742, 29.4012504])
        np.testing.assert_almost_equal(k_sat / 1.0e9, ksat_ref)

    def test_gassmann2(self):
        k_brine, phie, k_min, k_dry, k_oil = self.setup_gassmann()
        k_sat = gassmann(k_dry, phie, k_brine, k_min)
        k_sat2 = gassmann2(k_sat, k_brine, k_oil, phie, k_min)
        ksat2_ref = np.array([36.6, 15.7843355, 28.4290396])
        np.testing.assert_almost_equal(k_sat2 / 1.0e9, ksat2_ref)

    def test_gassmann_dry(self):
        k_brine, phie, k_min, k_dry, _ = self.setup_gassmann()
        k_sat = gassmann(k_dry, phie, k_brine, k_min)
        k_dry2 = gassmann_dry(k_sat, phie, k_brine, k_min)
        k_dry2_ref = np.array([36.6e9, 15.0e9, 28.0e9]) / 1e9
        np.testing.assert_almost_equal(k_dry2 / 1.0e9, k_dry2_ref)



import numpy as np
import numpy.typing as npt


def pressure(
    rho: npt.NDArray[np.float64],
    tvd_msl: npt.NDArray[np.float64],
    water_depth: float,
    p_form: float,
    tvd_p_form: float,
    n: float,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Function to estimate overburden pressure and vertical effective stress (lithostatic pressure)
    based on density.

    Parameters
    ----------
    rho : np.ndarray
        Density log [kg/m3].
    tvd_msl : np.ndarray
        Vertical depth log [m].
    water_depth : float
        Down to this point the difference between formation pressure and overburden pressure shall be zero [m].
    p_form : float
        Formation pressure [Pa].
    tvd_p_form : float
        Depth of formation pressure point [m].
    n: float
        Biot coefficient [unitless].

    Returns
    -------
    tuple
        p_eff, p_lith : np.ndarray.
        p_eff [Pa] - effective pressure,
        p_lith [Pa] - overburden pressure.
    """

    # Standard brine density with salinity 40000 ppm, 2 MPa and 4 deg. C
    rho_brine = 1.03e3
    # Gravity constant
    g = 9.80665
    # Input log length
    log_length = len(rho)

    # Allocate output logs
    p_eff = np.ones(log_length) * np.nan
    p_lith = np.ones(log_length) * np.nan

    # Does the depth log start at, above or below the sea bottom? We want it to
    # start at sea bottom, so we get rid of values above this point.
    # Also, as the tvd log normally is calculated by calling application, there can be
    # undefined values, flagged as negative or NaN
    idx = np.ones_like(rho, dtype=bool)
    idx_inf_nan = np.any(
        [np.isnan(rho), np.isinf(rho), np.isnan(tvd_msl), np.isinf(tvd_msl)], axis=0
    )
    idx[idx_inf_nan] = False
    idx[~idx_inf_nan] = np.logical_and(
        tvd_msl[~idx_inf_nan] >= water_depth, rho[~idx_inf_nan] > 0
    )

    # We need a starting point for pressure at water bottom - use a standard
    # density value
    p_wb = g * rho_brine * water_depth

    # p_form is a single value. No functionality is added at present to handle p_form in log version
    # Find which depth in the log that matches the calibration point best
    calib_point = np.argmin(abs(tvd_msl[idx] - tvd_p_form))

    # Overburden pressure
    dz = np.diff(tvd_msl[idx])
    # Append one sample to match the density log length
    dz = np.append(dz, dz[-1])

    # Find a starting point for the litho pressure at the start of the depth log
    # Density at water bottom is assumed to be for a sand with 40% porosity. Take the average
    # of this and the first observation of density log
    ave_dens = 0.5 * ((2650 * 0.6 + rho_brine * 0.4) + rho[0])
    p_lith_start = ave_dens * g * (tvd_msl[0] - water_depth)

    # Estimate the overburden pressure as the gravity of the cumulative bulk
    # density plus the calculated starting point and the fluid pressure at sea bottom
    p_lith[idx] = np.cumsum(g * dz * rho[idx]) + p_lith_start + p_wb

    # Find the effective/differential pressure at the calibration point
    p_eff_calib = p_lith[idx][calib_point] - n * p_form
    # In the absence of any better alternative - make a linear interpolation
    # from zero at water bottom to the calculated effective pressure at the
    # calibration point
    p_eff[idx] = np.interp(tvd_msl[idx], [water_depth, tvd_p_form], [0, p_eff_calib])

    p_lith[idx] = np.interp(tvd_msl[idx], tvd_msl[idx], p_lith[idx])

    return p_eff, p_lith



import numpy as np
from numpy.random import default_rng

from rock_physics_open.equinor_utilities.classification_functions import lin_class


class TestLinClass:
    def test_lin_class(self):
        rg = default_rng(234769238476)
        obs = rg.random((11, 2))
        class_mean = np.array(
            [
                [0.4450828, 0.5032985],
                [0.4959657, 0.4961214],
                [0.5020109, 0.5090125],
            ]
        )
        class_id = np.array([1, 2, 3])
        lin_class_arr, lin_dist = lin_class(obs, class_mean, class_id)
        lin_class_ref = np.array([1, 3, 1, 2, 3, 1, 3, 3, 1, 3, 1])
        lin_dist_ref = np.array(
            [
                0.4926242,
                0.2636271,
                0.2291065,
                0.4904759,
                0.0985496,
                0.4212475,
                0.3646983,
                0.1860129,
                0.2648174,
                0.0640302,
                0.3429182,
            ]
        )
        np.testing.assert_almost_equal(lin_class_arr, lin_class_ref)
        np.testing.assert_almost_equal(lin_dist, lin_dist_ref)



from pathlib import Path

import numpy as np
import numpy.typing as npt
import pandas as pd

from rock_physics_open.equinor_utilities.machine_learning_utilities.run_regression import (
    run_regression,
)


def carbonate_pressure_model(
    rho_fluid: npt.NDArray[np.float64],
    vp_in_situ: npt.NDArray[np.float64],
    vs_in_situ: npt.NDArray[np.float64],
    rho_in_situ: npt.NDArray[np.float64],
    vp_fluid_sub: npt.NDArray[np.float64],
    vs_fluid_sub: npt.NDArray[np.float64],
    rho_fluid_sub: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    pres_overburden: npt.NDArray[np.float64],
    pres_formation: npt.NDArray[np.float64],
    pres_form_depleted: npt.NDArray[np.float64],
    vp_model: Path,
    vs_model: Path,
    model_path: Path,
    b_add_fluid_sub: bool = False,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Function for estimating relative changes in Vp and Vs as a function of formation pressure depletion for carbonates.
    The models that are used are based on laboratory plug measurements at close to in situ conditions,
    and with simulated formation pressure depletion. The training and validation data set is based on pre-salt
    carbonates from producing fields from the Brazilian continental shelf.

    Similar to other substitution routines, the estimated difference between in situ case and depletion case is added
    to observed logs.

    A fluid substitution is assumed to be run in advance of the pressure substitution, and the depletion effect is added
    to the substituted logs. In case there is no preceding fluid substitution, the in situ logs are presented for both
    the in situ and fluid substituted case.

    Parameters
    ----------
    rho_fluid : np.ndarray
        Fluid density [kg/m^3]
    vp_in_situ : np.ndarray
        In situ Vp [m/s]
    vs_in_situ : np.ndarray
        In situ Vs [m/s]
    rho_in_situ : np.ndarray
        In situ bulk density [kg/m^3]
    vp_fluid_sub : np.ndarray
        Fluid substituted Vp [m/s]
    vs_fluid_sub : np.ndarray
        Fluid substituted Vs [m/s]
    rho_fluid_sub : np.ndarray
        Fluid substituted bulk density [kg/m^3]
    phi : np.ndarray
        Porosity [fraction]
    pres_overburden : np.ndarray
        Overburden pressure [Pa]
    pres_formation : np.ndarray
        In situ formation pressure [Pa]
    pres_form_depleted : np.ndarray
        Depleted formation pressure [Pa]
    vp_model : Path
        Full name to neural network model for Vp
    vs_model : Path
        Full name to neural network model for Vs
    model_path : Path
        Path to model directory
    b_add_fluid_sub : bool
        Control whether effect of fluid substitution should be added to the final result

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        vp_pres_sub : pressure substituted p-velocity [m/s]
        vs_pres_sub : pressure substituted s-velocity [m/s]
        rho_pres_sub : fluid substituted density is returned as the pressure substitution does not change this [kg/m^3]
        ai_pres_sub : derived acoustic impedance [m/s x kg/m^3]
        vpvs_pres_sub : derived vp/vs-ratio [fraction]
    """

    # Plug measurements are reported or dry rock density, bulk density needs to be corrected for fluid effect
    rho_dry = rho_in_situ - rho_fluid * phi

    # Change unit to comply with laboratory units
    # Porosity in percent
    phi_percent = phi * 100.0
    # Pressure in MPa
    pres_overburden *= 1.0e-6
    pres_formation *= 1.0e-6
    pres_form_depleted *= 1.0e-6
    # Density in g/cm^3
    rho_dry *= 1.0e-3

    # Generate DataFrame with necessary inputs for in situ and depleted cases
    input_dict = {
        "PHIT": phi_percent,
        "VP": vp_in_situ,
        "VSX": vs_in_situ,
        "PEFF_in_situ": pres_overburden - pres_formation,
        "PEFF_depleted": pres_overburden - pres_form_depleted,
    }
    input_df = pd.DataFrame(input_dict)

    results_df = run_regression(
        inp_df=input_df,
        first_model_file_name=str(vp_model),
        second_model_file_name=str(vs_model),
        model_dir=str(model_path),
    )
    results_df.columns = ["vp_delta", "vs_delta"]

    if b_add_fluid_sub:
        start_vp = vp_fluid_sub
        start_vs = vs_fluid_sub
        start_rho = rho_fluid_sub
    else:
        start_vp = vp_in_situ
        start_vs = vs_in_situ
        start_rho = rho_in_situ

    vp_pres_sub = start_vp + results_df["vp_delta"].to_numpy()
    vs_pres_sub = start_vs + results_df["vs_delta"].to_numpy()
    rho_pres_sub = start_rho
    ai_pres_sub = vp_pres_sub * rho_pres_sub
    vpvs_pres_sub = vp_pres_sub / vs_pres_sub

    return vp_pres_sub, vs_pres_sub, rho_pres_sub, ai_pres_sub, vpvs_pres_sub



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from .gen_ternary_plot import ternary_plot
from .shale_prop_ternary import shale_prop_ternary
from .ternary_patches import ternary_patches


def run_ternary(
    quartz: npt.NDArray[np.float64],
    carb: npt.NDArray[np.float64],
    clay: npt.NDArray[np.float64],
    kero: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    misc: npt.NDArray[np.float64],
    misc_log_type: str,
    well_name: str,
    draw_figures: bool = True,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """Combined call to three different ternary plots used to describe unconventionals (shale) models.

    Parameters
    ----------
    quartz : np.ndarray
        Quartz volume fraction [fraction].
    carb : np.ndarray
        Carbonate volume fraction [fraction].
    clay : np.ndarray
        Clay volume fraction [fraction].
    kero : np.ndarray
        Kerogen volume fraction [fraction].
    phi : np.ndarray
        Porosity [fraction].
    misc : np.ndarray
        Property used for colour coding [unknown].
    misc_log_type : str
        Plot annotation of log used for colour coding.
    well_name : str
        Plot heading with well name.
    draw_figures : bool
        Decide if figures are drawn or not, default is True.

    Returns
    -------
    tuple
        lith_class, hard : (np.ndarray, np.ndarray).
        lith_class: lithology class [int], hardness [float].
    """
    matplotlib.use("TkAgg")

    lith_class = ternary_patches(
        quartz, carb, clay, kero, well_name, draw_figures=draw_figures
    )

    hard = shale_prop_ternary(
        quartz=quartz,
        carb=carb,
        clay=clay,
        kero=kero,
        phit=phi,
        col_code=misc,
        name_col_code=misc_log_type,
        well_name=well_name,
        draw_figures=draw_figures,
    )

    ternary_plot(
        data1=quartz,
        data2=carb,
        data3=clay,
        data4=kero,
        well_name=well_name,
        name_data1="Quartz",
        name_data2="Carbonate",
        name_data3="Clay",
        name_data4="Kerogen",
        draw_figures=draw_figures,
    )

    if draw_figures:
        plt.show()

    return lith_class, hard



from typing import cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import gen_utilities, std_functions


def contact_cement_model(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    frac_cem: float,
    phi_c: float,
    n: float,
    shear_red: float,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Contact cement model is a sandstone model where all variation in porosity is explained by variation in grain contact
    cement. This is leads to the stiffest transition from critical porosity to the mineral point.

    Mineral properties k_min, muMin, rho_min are effective properties. For mixtures of minerals, effective properties
    are calculated by Hashin-Shtrikman or similar.

    Cement properties kCem, muCem, rhoCem and cement fraction, the latter as a part of the whole volume.

    Fluid properties k_fl, rho_fl are in situ properties calculated by fluid models using reservoir properties.

    Critical porosity phiC is input to Dvorkin-Nur function. Coordination number n is normally set to a fixed value
    (default 9) but it is possible to  override this. Porosity phi is used in Hashin-Shtrikman mixing (together with
    phiC) and Gassmann saturation.

    Shear reduction parameter shearRed is used to account for tangential frictionless grain contacts.

    All inputs are assumed to be vectors of the same length.


    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    mu_min : np.ndarray
        Mineral shear modulus [Pa].
    rho_min : np.ndarray
        Mineral bulk density [kg/m^3].
    k_cem : np.ndarray
        Cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Cement shear modulus [Pa].
    rho_cem : np.ndarray
        Cement bulk density [kg/m^3].
    k_fl : np.ndarray
        Fluid bulk modulus [Pa].
    rho_fl : np.ndarray
        Fluid bulk density [kg/m^3].
    phi : np.ndarray
        Porosity [fraction].
    frac_cem : float
        Cement fraction [fraction].
    phi_c : float
        Critical porosity [fraction].
    n : float
        Coordination number [unitless].
    shear_red : float
        Shear reduction factor [fraction].

    Returns
    -------
    tuple
        vp, vs, rho, ai, vpvs  : np.ndarray
        vp [m/s] and vs [m/s], bulk density [kg/m^3], ai [m/s x kg/m^3], vpvs [ratio] of saturated rock.
    """

    # Identify porosity values that are above (phi_c - frac_cem), for which the model is not defined
    (
        idx_phi,
        (
            k_min,
            mu_min,
            rho_min,
            k_cem,
            mu_cem,
            rho_cem,
            k_fl,
            rho_fl,
            phi,
            _,
        ),
    ) = cast(
        tuple[npt.NDArray[np.bool_], list[npt.NDArray[np.float64]]],
        gen_utilities.filter_input_log(
            (
                k_min,
                mu_min,
                rho_min,
                k_cem,
                mu_cem,
                rho_cem,
                k_fl,
                rho_fl,
                phi,
                phi_c - frac_cem - phi,
            )
        ),
    )

    # Expand input parameters to arrays
    phi, frac_cem_, phi_c_, _, _ = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.dim_check_vector((phi, frac_cem, phi_c, n, shear_red)),
    )

    # At the zero-porosity point, all original porosity (critical porosity) is
    # filled with cement. The cement surrounds the original grains, so they
    # will be fraction 1 according to the geometrical interpretation of
    # Hashin-Shtrikman
    k_zero, mu_zero = std_functions.hashin_shtrikman_walpole(
        k1=k_cem,
        mu1=mu_cem,
        k2=k_min,
        mu2=mu_min,
        f1=phi_c_,
        bound="lower",
    )

    # Dry rock properties of high-porosity end member calculated with
    # Dvorkin-Nur equation. Cement is assumed to be evenly distributed on the
    # grains (scheme 2 in Dvorkin and Nur's original paper)

    k_cc, mu_cc = std_functions.dvorkin_contact_cement(
        frac_cem=frac_cem_,
        por0_sst=phi_c,
        mu0_sst=mu_zero,
        k0_sst=k_zero,
        mu0_cem=mu_cem,
        k0_cem=k_cem,
        vs_red=shear_red,
        c=n,
    )

    # Hashin-Shtrikman upper bound describes the dry rock property mixing from
    # mineral properties to high-end porosity.

    # Fraction of zero-porosity end member
    f1 = 1 - phi / (phi_c - frac_cem)

    k_dry, mu = std_functions.hashin_shtrikman_walpole(
        k1=k_zero,
        mu1=mu_zero,
        k2=k_cc,
        mu2=mu_cc,
        f1=f1,
        bound="upper",
    )

    # Saturated rock incompressibility is calculated with Gassmann
    k = std_functions.gassmann(
        k_dry=k_dry,
        por=phi,
        k_fl=k_fl,
        k_min=k_zero,
    )

    # Bulk density
    rho = phi * rho_fl + (1 - phi_c) * rho_min + (phi_c - phi) * rho_cem

    vp, vs, ai, vpvs = std_functions.velocity(k=k, mu=mu, rhob=rho)

    # Restore original array length
    vp, vs, rho, ai, vpvs = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.filter_output(idx_phi, (vp, vs, rho, ai, vpvs)),
    )

    return vp, vs, rho, ai, vpvs



import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions
from rock_physics_open.sandstone_models.constant_cement_models import (
    constant_cement_model,
)
from rock_physics_open.sandstone_models.friable_models import (
    CoordinateNumberFunction,
    friable_model,
)


def unresolved_constant_cement_sand_shale_model(
    k_min_sst: npt.NDArray[np.float64],
    mu_min_sst: npt.NDArray[np.float64],
    rho_min_sst: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    k_mud: npt.NDArray[np.float64],
    mu_mud: npt.NDArray[np.float64],
    rho_mud: npt.NDArray[np.float64],
    k_fl_sst: npt.NDArray[np.float64],
    rho_fl_sst: npt.NDArray[np.float64],
    k_fl_mud: npt.NDArray[np.float64],
    rho_fl_mud: npt.NDArray[np.float64],
    phi_sst: npt.NDArray[np.float64],
    phi_mud: npt.NDArray[np.float64],
    p_eff_mud: npt.NDArray[np.float64],
    shale_frac: npt.NDArray[np.float64],
    frac_cem: float,
    phi_c_sst: float,
    phi_c_mud: float,
    n_sst: float,
    coord_num_func_mud: CoordinateNumberFunction,
    n_mud: float,
    shear_red_sst: float,
    shear_red_mud: float,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Model for silisiclastic rocks with alternating layers of cemented sand and friable shale, and in which the layers
    are not resolved by the investigating signal. Backus average is used to calculate the anisotropic effect of the
    alternating layers.

    Parameters
    ----------
    k_min_sst : np.ndarray
        Sandstone matrix bulk modulus [Pa].
    mu_min_sst : np.ndarray
        Sandstone matrix shear modulus [Pa].
    rho_min_sst : np.ndarray
        Sandstone matrix bulk density [kg/m^3].
    k_cem : np.ndarray
        Sandstone cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Sandstone cement shear modulus [Pa].
    rho_cem : np.ndarray
        Sandstone cement bulk density [kg/m^3].
    k_mud : np.ndarray
        Shale bulk modulus [Pa].
    mu_mud : np.ndarray
        Shale shear modulus [Pa].
    rho_mud : np.ndarray
        Shale bulk density [kg/m^3].
    k_fl_sst : np.ndarray
        Fluid bulk modulus for sandstone fluid [Pa].
    rho_fl_sst : np.ndarray
        Fluid bulk density for sandstone fluid [kg/m^3].
    k_fl_mud : np.ndarray
        Fluid bulk modulus for shale fluid [Pa].
    rho_fl_mud : np.ndarray
        Fluid bulk density for shale fluid[kg/m^3].
    phi_sst : np.ndarray
        Sandstone porosity [fraction].
    phi_mud : np.ndarray
        Shale porosity [fraction].
    p_eff_mud : np.ndarray
        Effective pressure in mud [Pa].
    shale_frac : np.ndarray
        Shale fraction [fraction].
    frac_cem : float
        Cement volume fraction [fraction].
    phi_c_sst : float
        Critical porosity for sandstone [fraction].
    phi_c_mud : float
        Critical porosity for mud [fraction].
    n_sst : float
        Coordination number for sandstone [unitless].
    n_mud : float
        Coordination number for shale [unitless].
    coord_num_func_mud : str
        Indication if coordination number should be calculated from porosity or kept constant for shale.
    shear_red_sst : float
        Shear reduction factor for sandstone [fraction].
    shear_red_mud : float
        Shear reduction factor for mud [fraction].

    Returns
    -------
    tuple
        vpv, vsv, vph, vsh, rho : np.ndarray
        vertical p-wave velocity, vertical shear-wave velocity, horizontal p-wave velocity, horizontal shear-wave
        velocity (all [m/s]), bulk density [kg/m^3].
    """
    # Estimate the sand end member through the constant cement model
    vp_sst, vs_sst, rho_b_sst, _, _ = constant_cement_model(
        k_min=k_min_sst,
        mu_min=mu_min_sst,
        rho_min=rho_min_sst,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl=k_fl_sst,
        rho_fl=rho_fl_sst,
        phi=phi_sst,
        frac_cem=frac_cem,
        phi_c=phi_c_sst,
        n=n_sst,
        shear_red=shear_red_sst,
    )

    # Estimate the shale end member through the friable model
    vp_mud, vs_mud, rho_b_mud, _, _ = friable_model(
        k_min=k_mud,
        mu_min=mu_mud,
        rho_min=rho_mud,
        k_fl=k_fl_mud,
        rho_fl=rho_fl_mud,
        phi=phi_mud,
        p_eff=p_eff_mud,
        phi_c=phi_c_mud,
        coord_num_func=coord_num_func_mud,
        n=n_mud,
        shear_red=shear_red_mud,
    )

    # Calculate Backus average for the effective medium
    vpv, vsv, vph, vsh, rho = std_functions.backus_average(
        vp1=vp_sst,
        vs1=vs_sst,
        rho1=rho_b_sst,
        vp2=vp_mud,
        vs2=vs_mud,
        rho2=rho_b_mud,
        f1=1.0 - shale_frac,
    )

    return vpv, vsv, vph, vsh, rho



import os

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.shale_models import (
    multi_sca,
    self_consistent_approximation_model,
)

k1 = 36.8e9 * np.ones(20)
mu1 = 44.0e9 * np.ones(20)
rho1 = 2650 * np.ones(20)
k2 = 15.0e9 * np.ones(20)
mu2 = 7.5e9 * np.ones(20)
rho2 = 2680 * np.ones(20)
k3 = 71.0e9 * np.ones(20)
mu3 = 32.5e9 * np.ones(20)
rho3 = 2720 * np.ones(20)
frac1 = np.linspace(0.0, 1.0, 20)
frac_inc = np.linspace(0.0, 0.4, 20)
frac_inc_1 = np.linspace(0.3, 0.2, 20)
asp1 = 0.5 * np.ones(20)
asp2 = 0.85 * np.ones(20)
tol = 1.0e-6


def test_sca_model():
    args = self_consistent_approximation_model(
        k1, mu1, rho1, k2, mu2, rho2, frac1, asp1, asp2, tol
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_multi_sca_model():
    args = multi_sca(
        k1, mu1, rho1, asp1, frac1, k2, mu2, rho2, asp2, 1 - frac1, tol=tol
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



import os

import numpy as np
import numpy.typing as npt
import pandas as pd

from rock_physics_open.equinor_utilities.optimisation_utilities import gen_opt_routine
from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.t_matrix_models.curvefit_t_matrix_exp import (
    curvefit_t_matrix_exp,
)
from rock_physics_open.t_matrix_models.curvefit_t_matrix_min import (
    curve_fit_2_inclusion_sets,
)

from ..conftest import TESTDATA_DIR

inp_df = pd.read_csv(TESTDATA_DIR / "tmatrix_test_data.csv")

k_min = inp_df["K_MIN"].to_numpy() * 1.0e9
mu_min = inp_df["MU_MIN"].to_numpy() * 1.0e9
rho_min = inp_df["RHO_MIN"].to_numpy() * 1000.0
k_fl_orig = inp_df["K_FL"].to_numpy() * 1.0e9
rho_fl_orig = inp_df["RHO_FL"].to_numpy() * 1000.0
vp = inp_df["VP"].to_numpy()
vs = inp_df["VS"].to_numpy()
rho = inp_df["RHOB"].to_numpy() * 1000.0
phi = inp_df["PHIT"].to_numpy()
vsh = inp_df["VSH"].to_numpy()
tau = 1e-7 * np.ones_like(phi)
angle = 45 * np.ones_like(phi)
perm = 100.0 * np.ones_like(phi)
visco = 100.0 * np.ones_like(phi)
freq = 1000.0 * np.ones_like(phi)
def_vp_vs_ratio = 1.8 * np.ones_like(phi)


"""
Direct testing of the T Matrix optimisation has proven to be difficult due to numerical differences when running
on different versions of Python and different architectures. Even if the number of significant digits are reduced,
there can be instances where the least significant digit will vary. The solution to this is to split the testing
in two parts: one for the optimisation in general and one for a single run of the T Matrix part of it. A simple
polynomial function is used for testing the optimisation.
"""


def poly_function(
    x_data: npt.NDArray[np.float64],
    a: float,
    b: float,
    c: float,
    d: float,
    e: float,
) -> npt.NDArray[np.float64]:
    """
    Simple function to test the scipy optimisation function curve_fit

    Parameters
    ----------
    x : np.ndarray
        independent variable in polynomial
    a : float
        constant
    b : float
        first order coefficient
    c : float
        second order coefficient
    d : float
        third order coefficient
    e : float
        fourth order coefficient

    Returns
    -------
    y : np.ndarray
        polynomial value
    """
    return a + b * x_data + c * x_data**2 + d * x_data**3 + e * x_data**4


def test_optimisation_part():
    a = 0.25398574328
    b = 0.28536214953
    c = 0.98235978612
    d = 0.23987512467
    e = 0.45897987345

    x = np.linspace(0, 1, 18)
    par_init = 0.5 * np.ones(5)
    lower_bound = np.zeros(5)  # Number of parameters in estimation
    upper_bound = np.ones(5)
    y = poly_function(x, a, b, c, d, e)
    _, _, args = gen_opt_routine(
        opt_function=poly_function,
        x_data_orig=x,
        y_data=y,
        x_init=par_init,
        low_bound=lower_bound,
        high_bound=upper_bound,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_t_matrix_opt_params_petec():
    x_data = np.stack(
        (
            phi,
            k_min,
            mu_min,
            rho_min,
            k_fl_orig,
            rho_fl_orig,
            angle,
            perm,
            visco,
            tau,
            freq,
            def_vp_vs_ratio,
        ),
        axis=1,
    )
    # Arbitrary setting for T Matrix parameters
    par = [0.5, 0.5, 0.5, 0.2, 0.2]
    args = curve_fit_2_inclusion_sets(x_data, *par)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_t_matrix_opt_params_exp():
    x_data = np.stack(
        (
            phi,
            vsh,
            k_fl_orig,
            rho_fl_orig,
            angle,
            perm,
            visco,
            tau,
            freq,
            def_vp_vs_ratio,
        ),
        axis=1,
    )
    # Arbitrary setting for T Matrix parameters
    par = [0.5, 0.5, 0.5, 0.30, 0.7, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]
    args = curvefit_t_matrix_exp(x_data, *par)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



from pathlib import Path

import pytest

from rock_physics_open.equinor_utilities.optimisation_utilities import (
    opt_param_to_ascii,
)


def test_opt_param_to_ascii_display_petec(
    monkeypatch: pytest.MonkeyPatch, data_dir: Path
):
    monkeypatch.chdir(data_dir)
    in_file = data_dir.joinpath("petec_opt_param_test.pkl")
    try:
        opt_param_to_ascii(in_file, display_results=False)
    except (ValueError, IOError):
        raise ValueError(f"Not possible to read input file {in_file}")
    out_file = data_dir.joinpath("petec_opt_param.txt")
    try:
        opt_param_to_ascii(in_file, display_results=False, out_file=out_file)
    except (ValueError, IOError):
        raise ValueError(f"Not possible to write output file {out_file}")


def test_opt_param_to_ascii_display_exp(
    monkeypatch: pytest.MonkeyPatch, data_dir: Path
):
    monkeypatch.chdir(data_dir)
    in_file = data_dir.joinpath("exp_opt_param_test.pkl")
    try:
        opt_param_to_ascii(in_file, display_results=False)
    except (ValueError, IOError):
        raise ValueError(f"Not possible to read input file {in_file}")
    out_file = data_dir.joinpath("exp_opt_param.txt")
    try:
        opt_param_to_ascii(in_file, display_results=False, out_file=out_file)
    except (ValueError, IOError):
        raise ValueError(f"Not possible to write output file {out_file}")



import warnings

import numpy as np
import numpy.typing as npt

from .dead_oil_density import dead_oil_density
from .dead_oil_velocity import dead_oil_velocity
from .live_oil_density import live_oil_density
from .live_oil_velocity import live_oil_velocity
from .oil_bubble_point import bp_standing


def oil_properties(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    rho0: npt.NDArray[np.float64],
    gas_oil_ratio: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    :param temperature: Temperature [°C] of oil.
    :param pressure: Formation pressure [Pa] of oil
    :param rho0: Density of the oil without dissolved gas at 15.6 degrees Celsius and
                 atmospheric pressure. [kg/m^3]
    :param gas_oil_ratio: The volume ratio of gas to oil [l/l]
    :param gas_gravity: Gas Gravity, molar mass of gas relative to air molar mas.
    :return: vel_oil [m/s], den_oil [kg/m^3], k_oil [Pa]
    """
    # Since live_oil with gas_oil_ratio=0.0 is not equal to dead oil
    # we use an apodization function to interpolate between the two

    def triangular_window(x: npt.NDArray[np.float64], length: int = 2):
        """
        A triangular window function around the origin, 1.0 at x=0.0, linear
        and 0.0 outside the window.
        :param length: total length of the window, ie., function is nonzero in
            [-length/2, length/2].
        :param x: numpy array containing x'es to evaluate the window at
        :return: value of window function at x.
        """
        x = np.asarray(x)  # Ensure x is a numpy array
        window = np.clip((np.abs(x) - length / 2) / (length / 2), 0, 1)
        return 1 - window

    loil_vel, loil_den = live_oil(
        temperature=temperature,
        pressure=pressure,
        reference_density=rho0,
        gas_oil_ratio=gas_oil_ratio,
        gas_gravity=gas_gravity,
    )
    doil_vel, doil_den = dead_oil(
        temperature=temperature,
        pressure=pressure,
        reference_density=rho0,
    )
    window = triangular_window(gas_oil_ratio)
    den_oil = doil_den * window + (1 - window) * loil_den
    vel_oil = doil_vel * window + (1 - window) * loil_vel
    k_oil = vel_oil**2 * den_oil
    return vel_oil, den_oil, k_oil


def dead_oil(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    reference_density: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    :param reference_density: Density of the oil without dissolved gas
        at 15.6 degrees Celsius and atmospheric pressure. [kg/m^3]
    :param gas_oil_ratio: The volume ratio of gas to oil [l/l]
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param pressure: Formation pressure [Pa] of oil
    :param temperature: Temperature [°C] of oil.
    :return: dead_oil_density [kg/m^3], dead_oil_velocity [m/s]
    """
    dead_oil_den = dead_oil_density(
        temperature=temperature,
        pressure=pressure,
        reference_density=reference_density,
    )
    dead_oil_vel = dead_oil_velocity(
        temperature=temperature,
        pressure=pressure,
        reference_density=reference_density,
    )
    return dead_oil_vel, dead_oil_den


def live_oil(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    reference_density: npt.NDArray[np.float64],
    gas_oil_ratio: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    :param reference_density: Density of the oil without dissolved gas
        at 15.6 degrees Celsius and atmospheric pressure. [kg/m^3]
    :param gas_oil_ratio: The volume ratio of gas to oil [l/l]
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param pressure: Formation pressure [Pa] of oil
    :param temperature: Temperature [°C] of oil.
    :return: live_oil_density , live_oil_velocity
    """
    if np.any(
        pressure
        < bp_standing(reference_density, gas_oil_ratio, gas_gravity, temperature)
    ):
        warnings.warn(
            "Formation pressure is below bubble point of oil, estimated elastic properties can be inaccurate",
            stacklevel=1,
        )
    live_oil_den = live_oil_density(
        temperature=temperature,
        pressure=pressure,
        reference_density=reference_density,
        gas_oil_ratio=gas_oil_ratio,
        gas_gravity=gas_gravity,
    )
    live_oil_vel = live_oil_velocity(
        temperature=temperature,
        pressure=pressure,
        reference_density=reference_density,
        gas_oil_ratio=gas_oil_ratio,
        gas_gravity=gas_gravity,
    )
    return (
        live_oil_vel,
        live_oil_den,
    )


def oil_viscosity(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    reference_density: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Calculate dead oil viscosity. If dissolved gas is present in the oil, the reference density
    should be substituted by live oil density.

    Equations 25a, 25b, 26a & 26b in Batzle and Wang 1992

    Based on Beggs and Robinson 1975

    :param temperature: Temperature [°C] of oil
    :param pressure: Formation pressure [Pa] of oil
    :param reference_density: Density of the oil without dissolved gas
    """
    # Change unit in pressure to MPa
    pressure_mpa = pressure / 1.0e6
    # Change unit in density to g/cc
    density_gcc = reference_density / 1000.0

    y_factor = 10 ** (5.693 - 2.863 / density_gcc)
    eta_t = -1.0 + 10 ** (0.505 * y_factor * (17.8 + temperature) ** -1.163)
    i_factor = 10 ** (
        18.6 * (0.1 * np.log10(eta_t) + (np.log10(eta_t) + 2) ** -0.1 - 0.985)
    )
    return eta_t + 0.145 * pressure_mpa * i_factor



import pickle
from typing import Any, Self, final

import numpy as np
from typing_extensions import override

from .base_pressure_model import BasePressureModel


@final
class SigmoidalPressureModel(BasePressureModel):
    """
    Sigmoidal pressure sensitivity model for velocity prediction.

    Uses nested sigmoid functions: velocity amplitude varies with porosity,
    and velocity varies sigmoidally with effective pressure using the amplitude.

    Input format (n,3): [porosity, p_eff_in_situ, p_eff_depleted]

    The model applies two sigmoid transformations:
    1. Porosity -> velocity amplitude using phi_model parameters
    2. Effective pressure -> velocity using p_eff_model parameters and amplitude
    """

    def __init__(
        self,
        phi_amplitude: float,
        phi_median_point: float,
        phi_x_scaling: float,
        phi_bias: float,
        p_eff_median_point: float,
        p_eff_x_scaling: float,
        p_eff_bias: float,
        model_max_pressure: float | None = None,
        description: str = "",
    ):
        """
        Initialize sigmoidal pressure model.

        Parameters
        ----------
        phi_amplitude : float
            Amplitude parameter for porosity sigmoid [m/s].
        phi_median_point : float
            Median point for porosity sigmoid [fraction].
        phi_x_scaling : float
            X-scaling parameter for porosity sigmoid [unitless].
        phi_bias : float
            Bias parameter for porosity sigmoid [m/s].
        p_eff_median_point : float
            Median point for pressure sigmoid [Pa].
        p_eff_x_scaling : float
            X-scaling parameter for pressure sigmoid [1/Pa].
        p_eff_bias : float
            Bias parameter for pressure sigmoid [m/s].
        model_max_pressure : float | None
            Maximum pressure for predict_max method [Pa].
        description : str
            Model description.
        """
        super().__init__(model_max_pressure, description)
        # Porosity model parameters
        self._phi_amplitude = phi_amplitude
        self._phi_median_point = phi_median_point
        self._phi_x_scaling = phi_x_scaling
        self._phi_bias = phi_bias
        # Pressure model parameters
        self._p_eff_median_point = p_eff_median_point
        self._p_eff_x_scaling = p_eff_x_scaling
        self._p_eff_bias = p_eff_bias

    @property
    def phi_amplitude(self) -> float:
        """Porosity sigmoid amplitude."""
        return self._phi_amplitude

    @property
    def phi_median_point(self) -> float:
        """Porosity sigmoid median point."""
        return self._phi_median_point

    @property
    def phi_x_scaling(self) -> float:
        """Porosity sigmoid x-scaling."""
        return self._phi_x_scaling

    @property
    def phi_bias(self) -> float:
        """Porosity sigmoid bias."""
        return self._phi_bias

    @property
    def p_eff_median_point(self) -> float:
        """Pressure sigmoid median point."""
        return self._p_eff_median_point

    @property
    def p_eff_x_scaling(self) -> float:
        """Pressure sigmoid x-scaling."""
        return self._p_eff_x_scaling

    @property
    def p_eff_bias(self) -> float:
        """Pressure sigmoid bias."""
        return self._p_eff_bias

    @override
    def validate_input(self, inp_arr: np.ndarray) -> np.ndarray:
        """
        Validate input for sigmoidal model.

        Parameters
        ----------
        inp_arr : np.ndarray
            Input array to validate.

        Returns
        -------
        np.ndarray
            Validated input array.

        Raises
        ------
        ValueError
            If input format is invalid.
        """
        if not isinstance(inp_arr, np.ndarray):  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
            raise ValueError("Input must be numpy ndarray.")  # pyright: ignore[reportUnreachable] | Kept for backward compatibility
        if inp_arr.ndim != 2 or inp_arr.shape[1] != 3:
            raise ValueError(
                "Input must be (n,3): [porosity, p_eff_in_situ, p_eff_depleted]"
            )
        return inp_arr

    def _sigmoid_phi(self, phi: np.ndarray) -> np.ndarray:
        """
        Calculate velocity amplitude from porosity using sigmoid function.

        Parameters
        ----------
        phi : np.ndarray
            Porosity values [fraction].

        Returns
        -------
        np.ndarray
            Velocity amplitude values [m/s].
        """
        return (
            self._phi_amplitude
            / (1 + np.exp(-self._phi_x_scaling * (phi - self._phi_median_point)))
            + self._phi_bias
        )

    def _sigmoid_p_eff(self, p_eff: np.ndarray, amplitude: np.ndarray) -> np.ndarray:
        """
        Calculate velocity from effective pressure using sigmoid function with amplitude.

        Parameters
        ----------
        p_eff : np.ndarray
            Effective pressure values [Pa].
        amplitude : np.ndarray
            Velocity amplitude values [m/s].

        Returns
        -------
        np.ndarray
            Velocity values [m/s].
        """
        return (
            amplitude
            / (1 + np.exp(-self._p_eff_x_scaling * (p_eff - self._p_eff_median_point)))
            + self._p_eff_bias
        )

    @override
    def predict_abs(self, inp_arr: np.ndarray, case: str = "in_situ") -> np.ndarray:
        """
        Calculate absolute velocity for specified pressure case.

        Parameters
        ----------
        inp_arr : np.ndarray
            Validated input array (n,3).
        case : str
            Pressure case: "in_situ" or "depleted".

        Returns
        -------
        np.ndarray
            Velocity values [m/s].
        """
        arr = self.validate_input(inp_arr)

        phi = arr[:, 0]
        p_in_situ = arr[:, 1]
        p_depleted = arr[:, 2]

        # Calculate velocity amplitude from porosity
        velocity_amplitude = self._sigmoid_phi(phi)

        # Select pressure based on case
        p_eff = p_in_situ if case == "in_situ" else p_depleted

        # Calculate velocity from effective pressure and amplitude
        return self._sigmoid_p_eff(p_eff, velocity_amplitude)

    @override
    def todict(self) -> dict[str, Any]:
        """Convert model to dictionary."""
        return {
            "phi_amplitude": self._phi_amplitude,
            "phi_median_point": self._phi_median_point,
            "phi_x_scaling": self._phi_x_scaling,
            "phi_bias": self._phi_bias,
            "p_eff_median_point": self._p_eff_median_point,
            "p_eff_x_scaling": self._p_eff_x_scaling,
            "p_eff_bias": self._p_eff_bias,
            "model_max_pressure": self._model_max_pressure,
            "description": self._description,
        }

    @override
    @classmethod
    def load(cls, file: str | bytes) -> Self:
        """Load sigmoidal model from pickle file."""
        with open(file, "rb") as f_in:
            d = pickle.load(f_in)

        return cls(
            phi_amplitude=d["phi_amplitude"],
            phi_median_point=d["phi_median_point"],
            phi_x_scaling=d["phi_x_scaling"],
            phi_bias=d["phi_bias"],
            p_eff_median_point=d["p_eff_median_point"],
            p_eff_x_scaling=d["p_eff_x_scaling"],
            p_eff_bias=d["p_eff_bias"],
            model_max_pressure=d["model_max_pressure"],
            description=d["description"],
        )



from typing import cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import gen_utilities, std_functions
from rock_physics_open.sandstone_models.constant_cement_models import (
    constant_cement_model,
)
from rock_physics_open.sandstone_models.friable_models import (
    CoordinateNumberFunction,
    friable_model,
)


def cemented_shaly_sand_sandy_shale_model(
    k_sst: npt.NDArray[np.float64],
    mu_sst: npt.NDArray[np.float64],
    rho_sst: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    k_mud: npt.NDArray[np.float64],
    mu_mud: npt.NDArray[np.float64],
    rho_mud: npt.NDArray[np.float64],
    k_fl_sst: npt.NDArray[np.float64],
    rho_fl_sst: npt.NDArray[np.float64],
    k_fl_mud: npt.NDArray[np.float64],
    rho_fl_mud: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    p_eff_mud: npt.NDArray[np.float64],
    shale_frac: npt.NDArray[np.float64],
    frac_cem: npt.NDArray[np.float64],
    phi_c_sst: float,
    n_sst: float,
    shear_red_sst: float,
    phi_c_mud: float,
    phi_intr_mud: float,
    coord_num_func_mud: CoordinateNumberFunction,
    n_mud: float,
    shear_red_mud: float,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Model for mixing of cemented sand and friable shale.

    It is no point to use this model to calculate the shale response only,
    in that case Friable Model with shale parameters does the job

    The shale fluid should be brine.

    Shale fraction shaleFrac is in the range 0 to 1. For shaleFrac = 0 we
    have a pure sand end member with phi =< phi_c for sand. For shaleFrac = 1
    we have pure shale with phi = intrinsic porosity. For shaleFrac < phi_c
    the model is on the shaly sand trend, for shaleFrac > phi_c it is on the
    sandy shale trend.

    Parameters
    ----------
    k_sst : np.ndarray
        Sandstone matrix bulk modulus [Pa].
    mu_sst : np.ndarray
        Sandstone matrix shear modulus [Pa].
    rho_sst : np.ndarray
        Sandstone matrix bulk density [kg/m^3].
    k_cem : np.ndarray
        Sandstone cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Sandstone cement shear modulus [Pa].
    rho_cem : np.ndarray
        Sandstone cement bulk density [kg/m^3].
    k_mud : np.ndarray
        Shale bulk modulus [Pa].
    mu_mud : np.ndarray
        Shale shear modulus [Pa].
    rho_mud : np.ndarray
        Shale bulk density [kg/m^3].
    k_fl_sst : np.ndarray
        Fluid bulk modulus for sandstone fluid [Pa].
    rho_fl_sst : np.ndarray
        Fluid bulk density for sandstone fluid [kg/m^3].
    k_fl_mud : np.ndarray
        Fluid bulk modulus for shale fluid [Pa].
    rho_fl_mud : np.ndarray
        Fluid bulk density for shale fluid[kg/m^3].
    phi : np.ndarray
        Total porosity [fraction].
    p_eff_mud : np.ndarray
        Effective pressure in mud [Pa].
    shale_frac : np.ndarray
        Shale fraction [fraction].
    frac_cem : np.ndarray
        Cement volume fraction [fraction].
    phi_c_sst : float
        Critical porosity for sandstone[fraction].
    n_sst : float
        Coordination number for sandstone [unitless].
    shear_red_sst : float
        Shear reduction factor for sandstone [fraction].
    phi_c_mud : float
        Critical porosity for mud [fraction].
    phi_intr_mud : float
        Intrinsic porosity for mud [fraction].
    coord_num_func_mud : str
        Indication if coordination number should be calculated from porosity or kept constant for shale.
    n_mud : float
        Coordination number for shale [unitless].
    shear_red_mud : float
        Shear reduction factor for mud [fraction].

    Returns
    -------
    tuple
        vp, vs, rho, ai, vpvs  : (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray)
        vp [m/s] and vs [m/s], bulk density [kg/m^3], ai [m/s x kg/m^3], vpvs [ratio] of saturated rock.
    """

    # Valid porosity values must not exceed (phi_c - frac_cem)
    # Assume that this only needs to be considered for the sandstone fraction
    (
        k_sst,
        mu_sst,
        rho_sst,
        k_cem,
        mu_cem,
        rho_cem,
        k_mud,
        mu_mud,
        rho_mud,
        k_fl_sst,
        rho_fl_sst,
        k_fl_mud,
        rho_fl_mud,
        phi,
        p_eff_mud,
        shale_frac,
        frac_cem_,
    ) = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.dim_check_vector(
            (
                k_sst,
                mu_sst,
                rho_sst,
                k_cem,
                mu_cem,
                rho_cem,
                k_mud,
                mu_mud,
                rho_mud,
                k_fl_sst,
                rho_fl_sst,
                k_fl_mud,
                rho_fl_mud,
                phi,
                p_eff_mud,
                shale_frac,
                frac_cem,
            )
        ),
    )
    (
        idx_phi,
        (
            k_sst,
            mu_sst,
            rho_sst,
            k_cem,
            mu_cem,
            rho_cem,
            k_mud,
            mu_mud,
            rho_mud,
            k_fl_sst,
            rho_fl_sst,
            k_fl_mud,
            rho_fl_mud,
            phi,
            p_eff_mud,
            shale_frac,
            frac_cem_,
            _,
            _,
        ),
    ) = cast(
        tuple[npt.NDArray[np.bool_], list[npt.NDArray[np.float64]]],
        gen_utilities.filter_input_log(
            (
                k_sst,
                mu_sst,
                rho_sst,
                k_cem,
                mu_cem,
                rho_cem,
                k_mud,
                mu_mud,
                rho_mud,
                k_fl_sst,
                rho_fl_sst,
                k_fl_mud,
                rho_fl_mud,
                phi,
                p_eff_mud,
                shale_frac,
                frac_cem_,
                phi_c_sst - frac_cem - phi,
                phi - phi_intr_mud,
            ),
            no_zero=False,
        ),
    )
    # Reduce range of porosity by frac_cem
    phi_c = phi_c_sst - frac_cem

    sandy_shale_idx = shale_frac > phi
    shaly_sand_idx = ~sandy_shale_idx

    # Fraction of silt in silt - shale trend
    frac_silt = (1 - shale_frac) / (1 - phi)
    # Fraction of sand in sand - silt trend
    frac_sand = 1 - shale_frac / phi

    # Shale properties for intrinsic porosity point NB!  The phi_intr_mud is
    # normally a parameter, but the assumption in Friable model is that it is a
    # log.  Make sure that it is of the same length as the other

    # Expand the needed variables from float to numpy array
    phi, phi_intr_mud_ = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.dim_check_vector((phi, phi_intr_mud)),
    )

    vp_sat_mud, vs_sat_mud, rho_b_mud, _, _ = friable_model(
        k_min=k_mud,
        mu_min=mu_mud,
        rho_min=rho_mud,
        k_fl=k_fl_mud,
        rho_fl=rho_fl_mud,
        phi=phi_intr_mud_,
        p_eff=p_eff_mud,
        phi_c=phi_c_mud,
        coord_num_func=coord_num_func_mud,
        n=n_mud,
        shear_red=shear_red_mud,
    )
    k_sat_mud, mu_sat_mud = std_functions.moduli(
        vp=vp_sat_mud, vs=vs_sat_mud, rhob=rho_b_mud
    )

    # Calculate cemented zero-porosity sand
    k_zero, mu_zero = std_functions.hashin_shtrikman_walpole(
        k1=k_cem,
        mu1=mu_cem,
        k2=k_sst,
        mu2=mu_sst,
        f1=frac_cem,
        bound="lower",
    )
    rho_zero = rho_cem * frac_cem + (1 - frac_cem) * rho_sst

    # Silt end member
    k_silt, mu_silt = std_functions.hashin_shtrikman_walpole(
        k1=k_sat_mud,
        mu1=mu_sat_mud,
        k2=k_zero,
        mu2=mu_zero,
        f1=phi_c,
    )
    rho_silt = rho_b_mud * phi + rho_zero * (1 - phi)

    # Estimate the sand end member through the constant cement model with phi =
    # phiCSst <= maybe dubious to expand parameter to vector with assumption
    # that kSst has the correct size

    vp_sat_sst, vs_sat_sst, rho_sat_sst, _, _ = constant_cement_model(
        k_min=k_sst,
        mu_min=mu_sst,
        rho_min=rho_sst,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl=k_fl_sst,
        rho_fl=rho_fl_sst,
        phi=phi,
        frac_cem=frac_cem,
        phi_c=phi_c_sst,
        n=n_sst,
        shear_red=shear_red_sst,
    )
    k_sat_sst, mu_sat_sst = std_functions.moduli(
        vp=vp_sat_sst, vs=vs_sat_sst, rhob=rho_sat_sst
    )

    k = np.ones(shale_frac.shape) * np.nan
    mu = np.ones(shale_frac.shape) * np.nan
    rhob = np.ones(shale_frac.shape) * np.nan

    # Points on sandy shale trend
    k[sandy_shale_idx], mu[sandy_shale_idx] = std_functions.hashin_shtrikman_walpole(
        k1=k_silt[sandy_shale_idx],
        mu1=mu_silt[sandy_shale_idx],
        k2=k_sat_mud[sandy_shale_idx],
        mu2=mu_sat_mud[sandy_shale_idx],
        f1=frac_silt[sandy_shale_idx],
    )

    rhob[sandy_shale_idx] = (
        rho_b_mud[sandy_shale_idx] * (1 - frac_silt[sandy_shale_idx])
        + rho_silt[sandy_shale_idx] * frac_silt[sandy_shale_idx]
    )

    # Points on shaly sand trend
    k[shaly_sand_idx], mu[shaly_sand_idx] = std_functions.hashin_shtrikman_walpole(
        k1=k_sat_sst[shaly_sand_idx],
        mu1=mu_sat_sst[shaly_sand_idx],
        k2=k_silt[shaly_sand_idx],
        mu2=mu_silt[shaly_sand_idx],
        f1=frac_sand[shaly_sand_idx],
    )

    rhob[shaly_sand_idx] = (
        (1 - phi[shaly_sand_idx]) * rho_zero[shaly_sand_idx]
        + phi[shaly_sand_idx] * rho_fl_sst[shaly_sand_idx]
    ) * frac_sand[shaly_sand_idx] + (1 - frac_sand[shaly_sand_idx]) * rho_silt[
        shaly_sand_idx
    ]

    vp, vs, ai, vpvs = std_functions.velocity(k=k, mu=mu, rhob=rhob)

    # Restore original length
    vp, vs, rhob, ai, vpvs = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.filter_output(idx_phi, (vp, vs, rhob, ai, vpvs)),
    )

    return vp, vs, rhob, ai, vpvs



import numpy as np
import numpy.typing as npt


def time_shift_pp(
    tvd: npt.NDArray[np.float64],
    vp_base: npt.NDArray[np.float64],
    vp_mon: npt.NDArray[np.float64],
    multiplier: int,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Cumulative time shift calculation for 4D case. According to Equinor standard
    the time shift is negative for an increase in velocity from base to monitor
    survey.

    Parameters
    ----------
    tvd : np.ndarray
        True vertical depth along wellbore [m].
    vp_base : np.ndarray
        Initial Vp [m/s].
    vp_mon : np.ndarray
        vp at time of monitor survey [m/s].
    multiplier : int
        Time shift multiplier.

    Returns
    -------
    tuple
        owt_pp_shift, twt_pp_shift : np.ndarray.
        owt_pp_shift: one way time shift [ms],
        twt_pp_shift: two way time shift [ms].

    Notes
    -----
    Original function by Sascha Bussat, Equinor
    Ported to Python by Harald Flesche, Equinor 2016.

    """
    dx = np.diff(tvd)
    dx = np.append(dx, dx[-1])

    time_base = np.cumsum(dx / vp_base)
    time_monitor = np.cumsum(dx / vp_mon)

    owt_pp_shift = (time_monitor - time_base) * 1000 * multiplier
    twt_pp_shift = 2 * owt_pp_shift

    return owt_pp_shift, twt_pp_shift


def time_shift_ps(
    tvd: npt.NDArray[np.float64],
    vp_base: npt.NDArray[np.float64],
    vp_mon: npt.NDArray[np.float64],
    vs_base: npt.NDArray[np.float64],
    vs_mon: npt.NDArray[np.float64],
    multiplier: int,
) -> npt.NDArray[np.float64]:
    """
    Cumulative time shift calculation for 4D case. According to Equinor standard
    the time shift is negative for an increase in velocity from base to monitor
    survey.

    Parameters
    ----------
    tvd : np.ndarray
        True vertical depth along wellbore [m].
    vp_base : np.ndarray
        Initial vp [m/s].
    vp_mon : np.ndarray
        vs at time of monitor survey [m/s].
    vs_base : np.ndarray
        Initial vp [m/s].
    vs_mon : np.ndarray
        vs at time of monitor survey [m/s].
    multiplier : int
        Time shift multiplier.

    Returns
    -------
    np.ndarray
        twt_ps_shift: two way time shift [ms]

    Notes
    -----
    Original function by Sascha Bussat, Equinor
    Ported to Python by Harald Flesche, Equinor 2016.

    """
    dx = np.diff(tvd)
    dx = np.append(dx, dx[-1])

    time_base_vp = np.cumsum(dx / vp_base)
    time_monitor_vp = np.cumsum(dx / vp_mon)

    time_base_vs = np.cumsum(dx / vs_base)
    time_monitor_vs = np.cumsum(dx / vs_mon)

    return (
        ((time_monitor_vp - time_base_vp) + (time_monitor_vs - time_base_vs))
        * 1000
        * multiplier
    )



import numpy as np
import numpy.typing as npt

from .dem import dem_model


def dem_model_dual_por(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    rho1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    rho2: npt.NDArray[np.float64],
    k3: npt.NDArray[np.float64],
    mu3: npt.NDArray[np.float64],
    rho3: npt.NDArray[np.float64],
    frac_inc: npt.NDArray[np.float64],
    frac_inc_1: npt.NDArray[np.float64],
    asp_1: npt.NDArray[np.float64],
    asp_2: npt.NDArray[np.float64],
    tol: float,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """Differential effective media model with two sets of inclusions.

    Parameters
    ----------
    k1 : np.ndarray
        Matrix bulk modulus [Pa].
    mu1 : np.ndarray
        Matrix shear modulus [Pa].
    rho1 : np.ndarray
        Matrix bulk density [kg/m^3].
    k2 : np.ndarray
        Inclusion 1 bulk modulus [Pa].
    mu2 : np.ndarray
        Inclusion 1 shear modulus [Pa].
    rho2 : np.ndarray
        Inclusion 1 bulk density [kg/m^3]
    k3 : np.ndarray
        Inclusion 2 bulk modulus [Pa].
    mu3 : np.ndarray
        Inclusion 2 shear modulus [Pa].
    rho3 : np.ndarray
        Inclusion 2 bulk density [kg/m^3].
    frac_inc : np.ndarray
        Total fraction of inclusions [fraction].
    frac_inc_1 : np.ndarray
        Fraction of inclusions belonging to type 1 [fraction].
    asp_1 : np.ndarray
        Aspect ratio of inclusion 1 [fraction].
    asp_2 : np.ndarray
        Aspect ratio of inclusion 2 [fraction].
    tol : float
        Tolerance of accuracy [unitless].

    Returns
    -------
    tuple
        k_dem_dual, mu_dem_dual, rhob_dem_dual : np.ndarray
        k_dem_dual: bulk modulus [Pa], mu_dem_dual: shear modulus [Pa], rhob_dem_dual: bulk density [kg/m^3].
    """
    # Include the Type 1 inclusions into the matrix first, then run again with inclusions Type 2
    k_dem1, mu_dem1, rhob_dem1 = dem_model(
        k1=k1,
        mu1=mu1,
        rho1=rho1,
        k2=k2,
        mu2=mu2,
        rho2=rho2,
        frac2=frac_inc * frac_inc_1,
        asp2=asp_1,
        tol=tol,
    )
    k_dem_dual, mu_dem_dual, rhob_dem_dual = dem_model(
        k1=k_dem1,
        mu1=mu_dem1,
        rho1=rhob_dem1,
        k2=k3,
        mu2=mu3,
        rho2=rho3,
        frac2=frac_inc * (1.0 - frac_inc_1),
        asp2=asp_2,
        tol=tol,
    )

    return k_dem_dual, mu_dem_dual, rhob_dem_dual



from typing import cast

import numpy as np
import numpy.typing as npt
import pandas as pd
from pandas.api.types import is_numeric_dtype
from sklearn.preprocessing import OneHotEncoder


def generate_dummy_vars(
    inp_frame: pd.DataFrame,
    class_var: str,
    ohe: OneHotEncoder | None = None,
) -> tuple[npt.NDArray[np.float64], int, npt.NDArray[np.str_]]:
    """
    From categorical variables generate a one-hot-encoder, i.e. each value in the categorical variable becomes a binary
    variable. See sklearn.preprocessing.OneHotEncoder.

    Parameters
    ----------
    inp_frame : pd.DataFrame
        Input data containing categorical variables.
    class_var : str
        Name of categorical variable.
    ohe : preprocessing.OneHotEncoder
        One-hot-encoder object.

    Returns
    -------
    dum_features, no_dummy_cols, dum_var_names : (np.ndarray, int, np.ndarray)
        dum_features: 2D array with transformed dummy variables, no_dummy_cols: number of columns in returned array,
        dum_var_names: automatically generated feature names.
    """

    if is_numeric_dtype(inp_frame[class_var]):
        # Make sure that the chosen indicator variable contains discrete values
        inp_frame = inp_frame.astype({class_var: "int32"})

    features_in = np.array(inp_frame[class_var]).reshape(-1, 1)

    if ohe is None:
        classes = features_in
        ohe = OneHotEncoder(categories="auto", sparse_output=False)
        _ = ohe.fit(classes)

    dum_features = cast(  # Casting since scikit-learn is not yet fully typed. `.transform` returns sparse matrix only if `sparse_output=True`.
        npt.NDArray[np.float64],
        ohe.transform(features_in),
    )
    no_dummy_cols = dum_features.shape[1]
    dum_var_names = ohe.get_feature_names_out()

    return dum_features, no_dummy_cols, dum_var_names



import warnings

import numpy as np
import numpy.typing as npt

from .pq import p_q_fcn


def kuster_toksoz_model(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    rho1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    rho2: npt.NDArray[np.float64],
    frac1: npt.NDArray[np.float64],
    asp2: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """Simplified Kuster-Toksoz model for single mineral inclusion with single aspect ratio.

    Parameters
    ----------
    k1 : np.ndarray
        Phase 1 bulk modulus [Pa].
    mu1 : np.ndarray
        Phase 1 shear modulus [Pa].
    rho1 : np.ndarray
        Phase 1 bulk density [kg/m^3].
    k2 : np.ndarray
        Phase 2 bulk modulus [Pa].
    mu2 : np.ndarray
        Phase 2 shear modulus [Pa].
    rho2 : np.ndarray
        Phase 2 bulk density [kg/m^3].
    frac1 : np.ndarray
        Fraction of phase 1 [fraction].
    asp2 : np.ndarray
        Aspect ratio for phase 2 inclusions [ratio].

    Returns
    -------
    tuple
        k_kt, mu_kt, rhob : np.ndarray
        effective media properties: k_kt: bulk modulus [Pa], mu_kt: shear modulus [Pa], rhob: bulk density [kg/m^3].
    """
    frac2 = 1.0 - frac1
    rhob = rho1 * frac1 + rho2 * frac2
    p, q = p_q_fcn(k1, mu1, k2, mu2, asp2)
    zeta = mu1 / 6.0 * (9.0 * k1 + 8.0 * mu1) / (k1 + 2.0 * mu1)

    k_kt = (k1 * (k1 + 4.0 / 3.0 * mu1) + 4.0 / 3.0 * mu1 * frac2 * (k2 - k1) * p) / (
        k1 + 4.0 / 3.0 * mu1 - frac2 * (k2 - k1) * p
    )
    mu_kt = (mu1 * (mu1 + zeta) + zeta * frac2 * (mu2 - mu1) * q) / (
        mu1 + zeta - frac2 * (mu2 - mu1) * q
    )

    # Non-physical situations can arise if there is too high volume fraction of inclusions with
    # low aspect ratio
    idx_neg = np.logical_or(k_kt < 0.0, mu_kt < 0.0)
    if np.any(idx_neg):
        k_kt[idx_neg] = np.nan
        mu_kt[idx_neg] = np.nan
        rhob[idx_neg] = np.nan
        warnings.warn(
            f"{__file__}: {np.sum(idx_neg)} non-physical solutions to Kuster-Toksöz equation"
        )

    return k_kt, mu_kt, rhob



from typing import cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import gen_utilities

# These routines are not finalised or used in any plugins yet
"""
	c11, c12, c13, c33, c44, c66 = c_ij_2_c_factors(cij)

	Transform a single stiffness tensor into components. VTI medium is assumed
	"""


def c_ij_2_c_factors(
    cij: npt.NDArray[np.float64],
) -> (
    tuple[
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
    ]
    | None
):
    """Transform a single stiffness tensor into components. VTI medium is assumed

    Parameters
    ----------
    cij : np.ndarray
            A 6x6 matrix.

    Returns
    -------
    tuple
            (c11, c12, c13, c33, c44, c66).
    """
    if not isinstance(cij, np.ndarray):  # pyright: ignore[reportUnnecessaryIsInstance]
        try:  # pyright: ignore[reportUnreachable]
            cij = np.array(cij, dtype=float)  # pyright: ignore[reportUnreachable]
        except ValueError:
            print("Input data can't be transformed into a NumPy array")  # pyright: ignore[reportUnreachable]
    try:
        num_samp = int(cij.size / 36)
        cij = cij.reshape((6, 6, num_samp))
        c11 = cij[0, 0, :].reshape(num_samp, 1)
        c12 = cij[0, 1, :].reshape(num_samp, 1)
        c13 = cij[0, 2, :].reshape(num_samp, 1)
        c33 = cij[2, 2, :].reshape(num_samp, 1)
        c44 = cij[3, 3, :].reshape(num_samp, 1)
        c66 = cij[5, 5, :].reshape(num_samp, 1)
        return c11, c12, c13, c33, c44, c66

    except ValueError:
        print("Input data is not a 6x6xN array")


def cfactors2cij(
    c11: npt.NDArray[np.float64],
    c12: npt.NDArray[np.float64],
    c13: npt.NDArray[np.float64],
    c33: npt.NDArray[np.float64],
    c44: npt.NDArray[np.float64],
    c66: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Transform individual stiffness factors to stiffness tensor 6x6x(number of samples).

    Parameters
    ----------
    c11, c12, c13, c33, c44, c66 : np.ndarray
            All 1-dimensional of same length.

    Returns
    -------
    np.ndarray
            A 6x6x(number of samples) stiffness tensor.
    """
    c11, c12, c13, c33, c44, c66 = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.dim_check_vector((c11, c12, c13, c33, c44, c66)),
    )

    num_samp = c11.shape[1]

    cij = np.zeros((6, 6, num_samp))
    cij[0, 0, :] = c11
    cij[0, 1, :] = c12
    cij[0, 2, :] = c13
    cij[1, 0, :] = c12
    cij[1, 1, :] = c11
    cij[1, 2, :] = c13
    cij[2, 0, :] = c13
    cij[2, 1, :] = c13
    cij[2, 1, :] = c33
    cij[3, 3, :] = c44
    cij[4, 4, :] = c44
    cij[5, 5, :] = c66

    return cij


def c_ij_2_thomsen(
    c: npt.NDArray[np.float64], rho: npt.NDArray[np.float64]
) -> (
    tuple[
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
    ]
    | None
):
    """Thomsen parameter for weak anisotropy.

    Parameters
    ----------
    c : np.ndarray
            A (log of or single instance of) 6x6 elastic tensor.
    rho : np.ndarray
            Density - log of same length as c.

    Returns
    -------
    tuple
            alpha, beta, gamma, delta, epsilon.
    """
    # C matrix should be 6x6
    if not isinstance(c, np.ndarray):  # pyright: ignore[reportUnnecessaryIsInstance]
        try:  # pyright: ignore[reportUnreachable]
            c = np.array(c, dtype=float)  # pyright: ignore[reportUnreachable]
        except ValueError:
            print("Input data can't be transformed into a NumPy array")  # pyright: ignore[reportUnreachable]
    try:
        num_samp = int(c.size / 36)
        c = c.reshape((6, 6, num_samp))
        rho = rho.reshape(num_samp, 1)
        alpha = np.sqrt(c[2, 2, :].reshape(num_samp, 1) / rho)
        beta = np.sqrt(c[3, 3, :].reshape(num_samp, 1) / rho)
        gamma = ((c[5, 5, :] - c[3, 3, :]) / (2 * c[3, 3, :])).reshape(num_samp, 1)
        epsilon = ((c[0, 0, :] - c[2, 2, :]) / (2 * c[2, 2, :])).reshape(num_samp, 1)
        delta = (
            ((c[0, 2, :] + c[3, 3, :]) ** 2 - (c[2, 2, :] - c[3, 3, :]) ** 2)
            / (2 * c[2, 2, :] * (c[2, 2, :] - c[3, 3, :]))
        ).reshape(num_samp, 1)

        return alpha, beta, epsilon, gamma, delta
    except ValueError:
        print("Input data is not a 6x6xN array")


def thomsen_2_c_ij(
    alpha: npt.NDArray[np.float64],
    beta: npt.NDArray[np.float64],
    gamma: npt.NDArray[np.float64],
    delta: npt.NDArray[np.float64],
    epsilon: npt.NDArray[np.float64],
    rho: npt.NDArray[np.float64],
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """Elastic stiffness. Assumptions:
            Thomsen's parameters apply for weak anisotropy in a transversely isotropic medium:

            c11 c12 c13  0   0   0

            c12 c11 c13  0   0   0

            c13 c13 c33  0   0   0

            0   0   0    c44 0   0

            0   0   0    0   c44 0

            0   0   0    0   0   c66

            Where c66 = 1/2(c11 - c12)

    Parameters
    ----------
    alpha, beta, gamma, delta, epsilon :
            Thomsen's parameters.
    rho :
            Bulk density.

    Returns
    -------
    tuple
            Elastic stiffness c11, c12, c13, c33, c44, c66.
    """
    alpha, beta, gamma, delta, epsilon = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.dim_check_vector((alpha, beta, gamma, delta, epsilon)),
    )

    c33 = rho * alpha**2
    c44 = rho * beta**2
    c11 = c33 * (1 + 2 * epsilon)
    c66 = c44 * (1 + 2 * gamma)
    c12 = c11 - 2 * c66
    c13 = np.sqrt(2 * c33 * (c33 - c44) * delta + (c33 - c44) ** 2) - c44

    return c11, c12, c13, c33, c44, c66



import numpy as np
import numpy.typing as npt

from .pq import p_q_fcn


def self_consistent_approximation_model(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    rho1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    rho2: npt.NDArray[np.float64],
    frac1: npt.NDArray[np.float64],
    asp1: npt.NDArray[np.float64],
    asp2: npt.NDArray[np.float64],
    tol: float,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    SCA - Effective elastic moduli using Berryman's Self-Consistent
    (Coherent Potential) Approximation method.

    Parameters
    ----------
    k1 : np.ndarray
        Bulk modulus of background matrix [Pa].
    mu1 : np.ndarray
        Shear modulus of background matrix [Pa].
    rho1 : np.ndarray
        Bulk density of background matrix [kg/m^3].
    k2 : np.ndarray
        Bulk modulus of inclusions [Pa].
    mu2 : np.ndarray
        Shear modulus of inclusions [Pa].
    rho2 : np.ndarray
        Bulk density of inclusions [kg/m^3].
    frac1 : np.ndarray
        Fraction of inclusions [fraction].
    asp1 : np.ndarray
        Aspect ratio of background matrix [ratio].
    asp2 : np.ndarray
        Aspect ratio of inclusions [ratio].
    tol: float
        Desired accuracy in the SCA iterations.

    Returns
    -------
    tuple
        k, mu, rho : np.ndarray
        k: effective medium bulk modulus [Pa], mu: effective medium shear modulus [Pa], rho: bulk density [kg/m^3].

    Comments
    --------
    Based on function by T. Mukerji, SRB, Stanford University, 1994.
    Ported to Python by Harald Flesche, Equinor 2015.
    """

    # Calculate fn and theta - independent of iteration
    f1 = frac1
    f2 = 1 - frac1

    k_sc = f1 * k1 + f2 * k2
    mu_sc = f1 * mu1 + f2 * mu2
    rhob = f1 * rho1 + f2 * rho2

    idx = np.logical_and(np.not_equal(f1, 0.0), np.not_equal(f1, 1.0))
    n_iter = 0
    k_new = np.zeros(np.sum(idx))
    delta = np.absolute(k_sc[idx] - k_new)
    # Express tolerance in terms of k1
    tol_ = tol * k1[idx]

    while np.any(delta > tol_) and (n_iter < 3000):
        p1, q1 = p_q_fcn(
            k=k_sc[idx],
            mu=mu_sc[idx],
            k2=k1[idx],
            mu2=mu1[idx],
            asp=asp1[idx],
        )
        p2, q2 = p_q_fcn(
            k=k_sc[idx],
            mu=mu_sc[idx],
            k2=k2[idx],
            mu2=mu2[idx],
            asp=asp2[idx],
        )

        k_new = (f1[idx] * k1[idx] * p1 + f2[idx] * k2[idx] * p2) / (
            f1[idx] * p1 + f2[idx] * p2
        )
        mu_new = (f1[idx] * mu1[idx] * q1 + f2[idx] * mu2[idx] * q2) / (
            f1[idx] * q1 + f2[idx] * q2
        )

        delta = np.absolute(k_sc[idx] - k_new)
        k_sc[idx] = k_new
        mu_sc[idx] = mu_new

        n_iter = n_iter + 1

    # If all inclusions or all matrix - substitute with inclusion mineral properties
    idx = frac1 == 1.0
    if np.any(idx):
        k_sc[idx] = k1[idx]
        mu_sc[idx] = mu1[idx]
    idx = frac1 == 0.0
    if np.any(idx):
        k_sc[idx] = k2[idx]
        mu_sc[idx] = mu2[idx]

    return k_sc, mu_sc, rhob



import numpy as np
import numpy.typing as npt

from .class_stats import gen_class_stats
from .mahal_class import mahal_class

NULL_CLASS = 0


def gen_two_step_class_stats(
    obs: npt.NDArray[np.float64],
    class_val: npt.NDArray[np.int64],
    thresh: float,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.int64],
    npt.NDArray[np.int64],
    npt.NDArray[np.float64],
]:
    """
    The observations are an n x m array, where n
    is the number of observations and m is the number of variables. With p
    classes the returned mean value will be an array of dimension p x m,
    covariance m x m x p and the class_id and prior probability p length vector.

    Generate statistics - mean, covariance and prior probability - for each
    class in the training data.	Run a mahalanobis classification, and exclude
    values that have distance above the threshold. Generate class statistics again
    and return them.

    Parameters
    ----------
    obs : np.ndarray
        An nxm array, where n is the number of samples and m is the number of variables.
    class_val : np.ndarray
        A p length vector, where p is the number of classes, containing class_id (integer numbers).
    thresh : float
        Unclassified threshold.

    Returns
    -------
    tuple
        (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray).
        class_mean, class_cov: statistics for each class after rejects;
        prior_prob, class_counts: based on number of observations in each class,
        find the prior probability of each class;
        class_id: label for each class;
        mahal_class_id: class id from mahalanobis classification.
    """
    mean_class_id, class_cov, _, _, class_id = gen_class_stats(obs, class_val)
    mahal_class_id = mahal_class(obs, mean_class_id, class_cov, class_id, thresh)[0]

    idx = mahal_class_id != NULL_CLASS
    mean_class_id, class_cov, prior_prob, class_counts, class_id = gen_class_stats(
        obs[idx], class_val[idx]
    )

    return mean_class_id, class_cov, prior_prob, class_counts, class_id, mahal_class_id



import matplotlib.path as mplpath
import numpy as np
import numpy.typing as npt


def poly_class(
    train_data: npt.NDArray[np.float64],
    polygons: npt.NDArray[np.float64],
    labels: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Points within the polygons are assigned to class labels. Point that do not
    fall within any polygon are set to NULL_CLASS.

    Parameters
    ----------

    train_data : np.ndarray
        Data points of two variables.
    polygons : np.ndarray
        Vertices of polygons in two-dimensional space.
    labels : np.ndarray
        Class label for each polygon.

    Returns
    -------
    np.ndarray
        Class id.
    """
    if len(labels) != len(polygons):
        raise ValueError("Number of labels are not matching number of polygons")

    # Create output variables
    idx_filtered = np.zeros(train_data.shape[0]).astype("bool")
    poly_class_id = np.zeros(train_data.shape[0]).astype("int")

    for i in range(len(polygons)):
        class_polygon = polygons[i]
        path = mplpath.Path(class_polygon)
        # Only points within the given polygon are used
        idx_poly = path.contains_points(train_data)
        poly_class_id[idx_poly] = labels[i]
        idx_filtered = np.logical_or(idx_filtered, idx_poly)
    # idx_filtered is no longer returned
    return poly_class_id



import os

import numpy as np
import pandas as pd

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.sandstone_models import (
    patchy_cement_model_cem_frac,
    patchy_cement_model_weight,
    patchy_cement_pressure_fluid_substitution,
)
from rock_physics_open.sandstone_models.friable_models import CoordinateNumberFunction

from ..conftest import TESTDATA_DIR

dataset = pd.read_csv(TESTDATA_DIR / "test_well.csv")

vp_old = dataset["VP"].to_numpy()
vs_old = dataset["VS"].to_numpy()
rho_b_old = dataset["RHOB"].to_numpy() * 1000.0
phi = dataset["PHIT"].to_numpy()
idx_high_phi = phi > 0.1
phi = phi[idx_high_phi]
vp_old = vp_old[idx_high_phi]
vs_old = vs_old[idx_high_phi]
rho_b_old = rho_b_old[idx_high_phi]
k_min = 136.8e9 * np.ones_like(phi)
mu_min = 44.0e9 * np.ones_like(phi)
rho_min = 2650 * np.ones_like(phi)
k_cem = 36.8e9 * np.ones_like(phi)
mu_cem = 44.0e9 * np.ones_like(phi)
rho_cem = 2650 * np.ones_like(phi)
k_fl_old = 0.8e9 * np.ones_like(phi)
rho_fl_old = 850 * np.ones_like(phi)
k_fl_new = 2.7e9 * np.ones_like(phi)
rho_fl_new = 1005 * np.ones_like(phi)
p_eff_old = 20.0e6 * np.ones_like(phi)
p_eff_new = 25.0e6 * np.ones_like(phi)
p_eff_low = 20.0e6 * np.ones_like(phi)
frac_cem_up = 0.10
frac_cem = 0.03
phi_c = 0.45
coord_num_func: CoordinateNumberFunction = "PorBased"
n = 8.0
shear_red = 0.3
weight_k = 0.6
weight_mu = 0.4


def test_patchy_cement_fluid_sub_model_no_change():
    args = patchy_cement_pressure_fluid_substitution(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl_old=k_fl_old,
        rho_fl_old=rho_fl_old,
        k_fl_new=k_fl_old,
        rho_fl_new=rho_fl_old,
        phi=phi,
        p_eff_old=p_eff_old,
        p_eff_new=p_eff_old,
        vp_old=vp_old,
        vs_old=vs_old,
        rho_b_old=rho_b_old,
        p_eff_low=p_eff_low,
        frac_cem_up=frac_cem_up,
        frac_cem=frac_cem,
        shear_red=shear_red,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        model_type="weight",
        phi_below_zero="disregard",
        phi_above_phi_c="disregard",
        k_sat_above_k_min="disregard",
        above_upper_bound="disregard",
        below_lower_bound="disregard",
    )
    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_patchy_cement_fluid_sub_model_weight():
    args = patchy_cement_pressure_fluid_substitution(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl_old=k_fl_old,
        rho_fl_old=rho_fl_old,
        k_fl_new=k_fl_new,
        rho_fl_new=rho_fl_new,
        phi=phi,
        p_eff_old=p_eff_old,
        p_eff_new=p_eff_new,
        vp_old=vp_old,
        vs_old=vs_old,
        rho_b_old=rho_b_old,
        p_eff_low=p_eff_low,
        frac_cem_up=frac_cem_up,
        frac_cem=frac_cem,
        shear_red=shear_red,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        model_type="weight",
        phi_below_zero="disregard",
        phi_above_phi_c="disregard",
        k_sat_above_k_min="disregard",
        above_upper_bound="disregard",
        below_lower_bound="disregard",
    )
    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_patchy_cement_fluid_sub_model_weight_snap():
    args = patchy_cement_pressure_fluid_substitution(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl_old=k_fl_old,
        rho_fl_old=rho_fl_old,
        k_fl_new=k_fl_new,
        rho_fl_new=rho_fl_new,
        phi=phi,
        p_eff_old=p_eff_old,
        p_eff_new=p_eff_new,
        vp_old=vp_old,
        vs_old=vs_old,
        rho_b_old=rho_b_old,
        p_eff_low=p_eff_low,
        frac_cem_up=frac_cem_up,
        frac_cem=frac_cem,
        shear_red=shear_red,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        model_type="weight",
        phi_below_zero="snap",
        phi_above_phi_c="snap",
        k_sat_above_k_min="snap",
        above_upper_bound="snap",
        below_lower_bound="snap",
    )
    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_patchy_cement_fluid_sub_model_cem_frac():
    args = patchy_cement_pressure_fluid_substitution(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl_old=k_fl_old,
        rho_fl_old=rho_fl_old,
        k_fl_new=k_fl_new,
        rho_fl_new=rho_fl_new,
        phi=phi,
        p_eff_old=p_eff_old,
        p_eff_new=p_eff_new,
        vp_old=vp_old,
        vs_old=vs_old,
        rho_b_old=rho_b_old,
        p_eff_low=p_eff_low,
        frac_cem_up=frac_cem_up,
        frac_cem=frac_cem,
        shear_red=shear_red,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        model_type="cement_fraction",
        phi_below_zero="disregard",
        phi_above_phi_c="disregard",
        k_sat_above_k_min="disregard",
        above_upper_bound="disregard",
        below_lower_bound="disregard",
    )
    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_patchy_cement_fluid_sub_model_cem_frac_snap():
    args = patchy_cement_pressure_fluid_substitution(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl_old=k_fl_old,
        rho_fl_old=rho_fl_old,
        k_fl_new=k_fl_new,
        rho_fl_new=rho_fl_new,
        phi=phi,
        p_eff_old=p_eff_old,
        p_eff_new=p_eff_new,
        vp_old=vp_old,
        vs_old=vs_old,
        rho_b_old=rho_b_old,
        p_eff_low=p_eff_low,
        frac_cem_up=frac_cem_up,
        frac_cem=frac_cem,
        shear_red=shear_red,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        model_type="cement_fraction",
        phi_below_zero="snap",
        phi_above_phi_c="snap",
        k_sat_above_k_min="snap",
        above_upper_bound="snap",
        below_lower_bound="snap",
    )
    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_patchy_cement_model_weight():
    args = patchy_cement_model_weight(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl=k_fl_old,
        rho_fl=rho_fl_old,
        phi=phi,
        p_eff=p_eff_old,
        frac_cem=frac_cem,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
        weight_k=weight_k,
        weight_mu=weight_mu,
    )
    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_patchy_cement_model_cem_frac():
    args = patchy_cement_model_cem_frac(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl=k_fl_old,
        rho_fl=rho_fl_old,
        phi=phi,
        p_eff=p_eff_old,
        frac_cem=frac_cem,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )
    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_patchy_cement_model_exceed_phi_extrapolate():
    frac_cem = 0.10
    phi_c = 0.40
    args = patchy_cement_model_cem_frac(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl=k_fl_old,
        rho_fl=rho_fl_old,
        phi=phi,
        p_eff=p_eff_old,
        frac_cem=frac_cem,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )
    # Set flag for extrapolation to True and assert that there are no NaN values in the output
    args = patchy_cement_model_cem_frac(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl=k_fl_old,
        rho_fl=rho_fl_old,
        phi=phi,
        p_eff=p_eff_old,
        frac_cem=frac_cem,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )
    assert not np.any(np.isnan(args[0]))

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



from typing import TypeVar, cast

import numpy as np
import numpy.typing as npt

_T = TypeVar("_T", npt.NDArray[np.float64], float)


def moduli(
    vp: _T,
    vs: _T,
    rhob: _T,
) -> tuple[_T, _T]:
    """
    Calculate isotropic moduli from velocity and density.

    Parameters
    ----------
    vp : np.ndarray or float
        Pressure wave velocity [m/s].
    vs : np.ndarray or float
        Shear wave velocity [m/s].
    rhob : np.ndarray or float
        Bulk density [kg/m^3].

    Returns
    -------
    tuple
        k, mu : (np.ndarray, np.ndarray) or (float, float).
        k: bulk modulus [Pa], mu: shear modulus [Pa].
    """
    mu = vs**2 * rhob
    k = vp**2 * rhob - 4 / 3 * mu

    return k, mu


def velocity(
    k: _T,
    mu: _T,
    rhob: _T,
) -> tuple[_T, _T, _T, _T]:
    """
    Calculate velocity, acoustic impedance and vp/vs ratio from elastic moduli and density.

    Parameters
    ----------
    k : np.ndarray, float
        Bulk modulus [Pa].
    mu : np.ndarray, float
        Shear modulus [Pa].
    rhob : np.ndarray, float
        Bulk density [kg/m^3].

    Returns
    -------
    tuple
        vp, vs, ai, vp_vs : np.ndarray.
        vp: pressure wave velocity [m/s], vs: shear wave velocity [m/s], ai: acoustic impedance [m/s x kg/m^3],
        vp_vs: velocity ratio [fraction].
    """
    vs = cast(_T, (mu / rhob) ** 0.5)
    vp = cast(_T, ((k + 4 / 3 * mu) / rhob) ** 0.5)
    ai = cast(_T, vp * rhob)
    vp_vs = cast(_T, vp / vs)

    return vp, vs, ai, vp_vs



import sys
from ctypes import c_double, c_int
from typing import Any, cast

import numpy as np
import numpy.ctypeslib as npct
from tmatrix import tmatrix_porosity_noscenario

from rock_physics_open.equinor_utilities import gen_utilities
from rock_physics_open.equinor_utilities.various_utilities.types import Array1D, Array2D

# Definition of input types for the T Matrix function
# this will affect the tests on the input data, dim_check_vector is therefore set up to
# return data on the specified format
array_1d_double = npct.ndpointer(dtype=c_double, ndim=1, flags="CONTIGUOUS")
array_1d_int = npct.ndpointer(dtype=c_int, ndim=1, flags="CONTIGUOUS")
array_2d_double = npct.ndpointer(dtype=c_double, ndim=2, flags="CONTIGUOUS")


def t_matrix_porosity_c_alpha_v(
    k_min: Array1D[np.float64],
    mu_min: Array1D[np.float64],
    rho_min: Array1D[np.float64],
    k_fl: Array1D[np.float64],
    rho_fl: Array1D[np.float64],
    phi: Array1D[np.float64],
    perm: Array1D[np.float64],
    visco: Array1D[np.float64],
    alpha: Array2D[np.float64],
    v: Array2D[np.float64],
    tau: Any,
    frequency: float,
    angle: float,
    frac_inc_con: float,
    frac_inc_ani: float,
) -> tuple[
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
]:
    """This function can be called directly from top level, but the present recommendation is to go though the run_t_matrix
    in order to check inputs. It is used directly from the optimisation functions for efficiency. This gives direct
    access to the C++ compiled library for T-Matrix.

    Parameters
    ----------
    k_min : np.ndarray
        N length array, mineral bulk modulus [Pa]
    mu_min: np.ndarray
        N length array, mineral shear modulus [Pa]
    rho_min: np.ndarray
        N length array, mineral density [kg/m^3]
    k_fl:np.ndarray
        N length array, fluid bulk modulus [Pa]
    rho_fl: np.ndarray
        N length array, fluid density [kg/m^3]
    phi : np.ndarray
        N length array, porosity
    perm: np.ndarray
        N length array, permeability [mD]
    visco: np.ndarray
        N length array, viscosity [cP]
    alpha: np.ndarray or float
        aspect ratios for inclusions
    v: np. ndarray or float
        fraction of porosity with given aspect ratio
    tau: float
        relaxation time
    frequency: float
        float single value, signal frequency [Hz]
    angle: float
        float single value, angle of symmetry plane (0 = HTI, 90 = VTI medium) [deg]
    frac_inc_con: np.ndarray or float
        float single value or array, fraction of inclusions that are connected
    frac_inc_ani: np.ndarray or float
        float single value or array, fraction of inclusions that are anisotropic

    Returns
    -------
    tuple
        Tuple of np.ndarrays. Vp: Vertical P-wave velocity [m/s], Vsv: Vertical polarity S-wave velocity [m/s],
        Vsh: Horizontal polarity S-wave velocity [m/s], Rhob [kg/m^3].
    """

    # ToDo: tau input is not used in the Calculo implementation of T-Matrix
    del tau

    # Make sure that what can be vectors are vectors of the same length.
    # frac_inc_con and frac_inc_ani can either be the same length as the log, be constants or have the same length as
    # alpha and v

    # frac_inc_con and frac_inc_ani must be of the same length
    frac_inc_con_, frac_inc_ani_ = (
        cast(  # casting since dim_check_vector typing is incomplete
            list[Array1D[np.float64]],
            gen_utilities.dim_check_vector((frac_inc_con, frac_inc_ani)),
        )
    )

    # test for frac_inc_con and frac_inc_ani being of the same length as the logs
    log_length = len(phi)
    if frac_inc_con_.shape[0] == log_length:
        (
            k_min,
            mu_min,
            rho_min,
            k_fl,
            rho_fl,
            phi,
            perm,
            visco,
            frac_inc_con_,
            frac_inc_ani_,
        ) = cast(  # casting since dim_check_vector typing is incomplete
            list[Array1D[np.float64]],
            gen_utilities.dim_check_vector(
                (
                    k_min,
                    mu_min,
                    rho_min,
                    k_fl,
                    rho_fl,
                    phi,
                    perm,
                    visco,
                    frac_inc_con_,
                    frac_inc_ani_,
                ),
                force_type=np.dtype("float64"),
            ),
        )
        frac_inc_length = log_length
    else:  # Single float value of frac_inc_con, frac_inc_ani or matching number of inclusions
        (
            k_min,
            mu_min,
            rho_min,
            k_fl,
            rho_fl,
            phi,
            perm,
            visco,
        ) = cast(  # casting since dim_check_vector typing is incomplete
            list[Array1D[np.float64]],
            gen_utilities.dim_check_vector(
                (k_min, mu_min, rho_min, k_fl, rho_fl, phi, perm, visco),
                np.dtype("float64"),
            ),
        )
        frac_inc_length = frac_inc_ani_.shape[0]

    # Create output array, gather mineral and fluid properties in 2D arrays
    out_arr = np.zeros((log_length, 4), dtype=float, order="C")
    min_prop = np.stack([k_min, mu_min, rho_min], axis=1)
    fl_prop = np.stack([k_fl, rho_fl, perm, visco], axis=1)

    # Make sure that alpha and v are of the same shape - more about length of alpha further down
    alpha_shape = alpha.shape
    alpha, v = cast(  # casting since dim_check_vector typing is incomplete
        list[Array2D[np.float64]],
        gen_utilities.dim_check_vector((alpha, v), force_type=np.dtype("float64")),
    )
    alpha = alpha.reshape(alpha_shape)
    v = v.reshape(alpha_shape)

    # Number of alphas can vary from sample to sample - not used here, regular number of alphas for all
    # samples. Need to declare the number of alphas per sample in an array. Alpha can also be a constant, in which case
    # it should be expanded to an array of log_length. Both alpha and v should be a 2D array
    if len(alpha) != log_length and len(alpha) < 5:
        # Interpret alpha as a vector of aspect ratios
        alpha_vec = (
            np.ones((log_length, len(alpha)), dtype=float, order="C") * alpha.flatten()
        )
        v_vec = np.ones((log_length, len(alpha)), dtype=float, order="C") * v.flatten()
        alpha = alpha_vec
        v = v_vec

    # Expect 2-dimensional input to t_mat_lib for alpha and v
    if alpha.ndim == 1:
        alpha = alpha.reshape((len(alpha), 1))
        v = v.reshape((len(alpha), 1))

    # Have to declare the number of alphas per sample, even if it is constant
    alpha_length_array = np.ones(log_length, dtype=c_int, order="C") * alpha.shape[1]
    alpha_length = alpha.shape[0]

    try:
        tmatrix_porosity_noscenario(
            out_np=out_arr,
            out_N=log_length,
            mineral_property_np=min_prop,
            fluid_property_np=fl_prop,
            phi_vector_np=phi,
            alpha_np=alpha,
            v_np=v,
            alpha_size_np=alpha_length_array,
            alpha_N=alpha_length,
            frequency=frequency,
            angle=angle,
            inc_con_np=frac_inc_con_,
            inc_ani_np=frac_inc_ani_,
            inc_con_N=frac_inc_length,
        )
    except ValueError:
        # Get more info in case this goes wrong
        raise TypeError(
            "tMatrix:t_matrix_porosity_c_alpha_v: {0}".format(str(sys.exc_info()))
        )

    vp = out_arr[:, 0]
    vsv = out_arr[:, 1]
    vsh = out_arr[:, 2]
    rhob = out_arr[:, 3]

    return vp, vsv, vsh, rhob



from typing import Literal, cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import gen_utilities, std_functions

CoordinateNumberFunction = Literal["PorBased", "ConstVal"]


def friable_model(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    p_eff: npt.NDArray[np.float64],
    phi_c: float,
    coord_num_func: CoordinateNumberFunction,
    n: float,
    shear_red: float,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Friable (non-cemented) sandstone model. All porosity variation is due to sorting, i.e. porosity filled with
    smaller grains. This model is pressure sensitive, but without asymptotic behaviour, which could be expected.

    The same model is also used for shale, with different set of parameters.

    Mineral properties k_min, mu_min, rho_min are effective properties. For
    mixtures of minerals, effective properties are calculated by
    Hashin-Shtrikman or similar.

    Fluid properties k_fl, rho_fl are in situ properties calculated by a fluid model using reservoir properties.

    Effective pressure p_eff is normally calculated as differential
    pressure, i.e. overburden pressure minus pore pressure.

    Critical porosity phi_c is input to Hertz-Mindlin function. Coordination
    number n is normally calculated from critical porosity, but it is
    possible to  override this through coord_num_func and parameter n.
    Porosity phi is used in Hashin-Shtrikman mixing (together with phi_c)
    and Gassmann saturation. Porosity values above the critical porosity are undefined and NaN is returned.

    Shear reduction parameter shearRed is used to account for tangential
    frictionless grain contacts.

    All inputs are assumed to be vectors of the same length. Single parameters
    are expanded to column vectors matching other inputs.

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    mu_min : np.ndarray
        Mineral shear modulus [Pa].
    rho_min : np.ndarray
        Mineral bulk density [kg/m^3].
    k_fl : np.ndarray
        Fluid bulk modulus [Pa].
    rho_fl : np.ndarray
        Fluid bulk density [kg/m^3].
    phi : np.ndarray
        Porosity [fraction].
    p_eff : np.ndarray
        Effective pressure [Pa].
    phi_c : float
        Critical porosity [fraction].
    coord_num_func : str
        Indication if coordination number should be calculated from porosity or kept constant.
    n : float
        Coordination number [unitless].
    shear_red : float
        Shear reduction factor [fraction].

    Returns
    -------
    tuple
        vp, vs, rho, ai, vpvs  : np.ndarray
        vp [m/s] and vs [m/s], bulk density [kg/m^3], ai [m/s x kg/m^3], vpvs [ratio] of saturated rock.
    """
    k_dry, mu = friable_model_dry(
        k_min=k_min,
        mu_min=mu_min,
        phi=phi,
        p_eff=p_eff,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )

    # Saturated rock incompressibility is calculated with Gassmann
    k = std_functions.gassmann(
        k_dry=k_dry,
        por=phi,
        k_fl=k_fl,
        k_min=k_min,
    )

    # Bulk density
    rhob = std_functions.rho_b(
        phi=phi,
        rho_f=rho_fl,
        rho_mat=rho_min,
    )

    vp, vs, ai, vpvs = std_functions.velocity(
        k=k,
        mu=mu,
        rhob=rhob,
    )

    return vp, vs, rhob, ai, vpvs


def friable_model_dry(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    p_eff: npt.NDArray[np.float64],
    phi_c: float,
    coord_num_func: CoordinateNumberFunction,
    n: float | None,
    shear_red: npt.NDArray[np.float64] | float,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Dry rock version of friable sandstone model.

    Mineral properties k_min, mu_min, are effective properties. For
    mixtures of minerals, effective properties are calculated by
    Hashin-Shtrikman or similar.

    Effective pressure p_eff is normally calculated as differential
    pressure, i.e. overburden pressure minus pore pressure.

    Critical porosity phi_c is input to Hertz-Mindlin function. Coordination
    number n is normally calculated from critical porosity, but it is
    possible to  override this through coordNumFunc and parameter n.
    Porosity phi is used in Hashin-Shtrikman mixing (together with phi_c).
    Porosity values above the critical porosity are undefined and NaN is returned.

    Shear reduction parameter shearRed is used to account for tangential
    frictionless grain contacts.

    All inputs are assumed to be vectors of the same length. Single parameters
    are expanded to column vectors matching other inputs.

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    mu_min : np.ndarray
        Mineral shear modulus [Pa].
    phi : np.ndarray
        Porosity [fraction].
    p_eff : np.ndarray
        Effective pressure [Pa].
    phi_c : float
        Critical porosity [fraction].
    coord_num_func : str
        Indication if coordination number should be calculated from porosity or kept constant.
    n : float | None
        Coordination number [unitless].
    shear_red : float
        Shear reduction factor [fraction].

    Returns
    -------
    tuple
        k, mu : np.ndarray.
        Bulk modulus k [Pa], shear modulus mu [Pa] of dry rock.
    """
    # Expand floats to arrays, check for equal length
    phi, phi_c_, shear_red_, k_min, mu_min, p_eff, n_ = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.dim_check_vector(
            (phi, phi_c, shear_red, k_min, mu_min, p_eff, n)
        ),
    )
    # Valid porosity values are less or equal to the critical porosity
    # Use filter_input_log to remove values that do not comply with this
    (
        idx_phi,
        (phi, phi_c_, shear_red_, k_min, mu_min, p_eff, _),
    ) = cast(
        tuple[npt.NDArray[np.bool_], list[npt.NDArray[np.float64]]],
        gen_utilities.filter_input_log(
            (phi, phi_c_, shear_red_, k_min, mu_min, p_eff, phi_c - phi)
        ),
    )

    # Dry rock properties of high-porosity end member calculated with
    # Hertz-Mindlin equation

    # Override coordination number calculation based on porosity
    if coord_num_func == "ConstVal":
        k_hm, mu_hm = std_functions.hertz_mindlin(
            k=k_min,
            mu=mu_min,
            phi_c=phi_c_,
            p=p_eff,
            shear_red=shear_red,
            coord=n_,
        )
    else:
        # Porosity based coordination number
        k_hm, mu_hm = std_functions.hertz_mindlin(
            k=k_min,
            mu=mu_min,
            phi_c=phi_c_,
            p=p_eff,
            shear_red=shear_red,
        )

    # Hashin-Shtrikman lower bound describes the dry rock property mixing from
    # mineral properties to high-end porosity.

    # Fraction of solid
    f1 = 1 - phi / phi_c

    k_dry, mu = std_functions.hashin_shtrikman_walpole(
        k_min, mu_min, k_hm, mu_hm, f1, bound="lower"
    )

    k_dry, mu = cast(
        list[npt.NDArray[np.float64]], gen_utilities.filter_output(idx_phi, (k_dry, mu))
    )

    return k_dry, mu



import os

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.shale_models import kuster_toksoz_model

k1 = 36.8e9 * np.ones(20)
mu1 = 44.0e9 * np.ones(20)
rho1 = 2650 * np.ones(20)
k2 = 11.0e9 * np.ones(20)
mu2 = 3e9 * np.ones(20)
rho2 = 2680 * np.ones(20)
frac1 = np.linspace(0.0, 1.0, 20)
asp_2 = 0.85 * np.ones(20)


def test_kus_tok_model():
    args = kuster_toksoz_model(k1, mu1, rho1, k2, mu2, rho2, frac1, asp_2)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array2D,
    Array3D,
    Array4D,
)

from .array_functions import array_inverse, array_matrix_mult
from .g_tensor import g_tensor_vec


def calc_td_vec(
    c0: Array3D[np.float64],
    i4: Array3D[np.float64],
    s_0: Array3D[np.float64],
    kd: Array4D[np.float64],
    alpha: Array2D[np.float64],
) -> Array4D[np.float64]:
    """Returns the dry t-matrix tensors (nx6x6x(numbers of empty cavities) matrix).

    Parameters
    ----------
    c0 : np.ndarray
        Stiffness tensor of the host material (nx6x6 matrix).
    i4 : np.ndarray
        An n x Identity matrix.
    s_0 : np.ndarray
        Inverse of stiffness tensor.
    kd : np.ndarray
        Dry K tensor of all the empty cavities (nx6x6x(numbers of empty cavities) matrix) see Agersborg et al.
        2009 for explanation.
    alpha : np.ndarray
        Aspect ratios of all the empty cavities (1x(numbers of empty cavities) vector).

    Returns
    -------
    np.ndarray
        Dry t-matrix tensors.
    """
    log_len = c0.shape[0]
    if alpha.ndim == 1 and alpha.shape[0] != c0.shape[0]:
        alpha = np.tile(alpha.reshape(1, alpha.shape[0]), (log_len, 1))
    alpha_len = alpha.shape[1]

    td = np.array(np.zeros((log_len, 6, 6, alpha_len)))

    for nc in range(alpha_len):
        g = g_tensor_vec(
            c0=c0,
            s_0=s_0,
            alpha=alpha[:, nc],
        )
        td[:, :, :, nc] = array_matrix_mult(
            array_inverse(g), array_matrix_mult(kd[:, :, :, nc], c0) - i4
        )

    return td



import os
import pickle
import sys
from pathlib import Path
from typing import Any, Literal, Required, TypedDict, cast

import numpy as np
import numpy.typing as npt
from scipy.optimize import curve_fit

from rock_physics_open.equinor_utilities.various_utilities.types import OptCallable


def curve_fit_wrapper(
    x_init: npt.NDArray[np.float64],
    opt_func: OptCallable,
    x_data: npt.NDArray[np.float64],
    y_data: npt.NDArray[np.float64],
    *args: Any,
    **opt_kwargs: Any,
) -> npt.NDArray[np.float64]:
    """Use in tests with scipy.optimize.minimize instead of curve_fit.

    Parameters
    ----------
    x_init : np.ndarray
        Initial guess for parameters.
    opt_func : callable
        Function to optimize.
    x_data : np.ndarray
        Input data to opt_func.
    y_data : np.ndarray
        Results that the optimisation should match.
    args :
        args opt_func.
    opt_kwargs :
        kwargs opt_func.

    Returns
    -------
    np.ndarray
        res values.
    """
    y_pred = opt_func(x_data, *x_init, *args, **opt_kwargs)
    return np.sum(np.sqrt((y_data - y_pred) ** 2))


def gen_opt_routine(
    opt_function: OptCallable,
    x_data_orig: npt.NDArray[np.float64],
    y_data: npt.NDArray[np.float64],
    x_init: npt.NDArray[np.float64],
    low_bound: npt.NDArray[np.float64],
    high_bound: npt.NDArray[np.float64],
    **opt_kwargs: Any,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    This function is a lean method for running optimisation with the given opt_function in curve_fit. Predicted values,
    residuals to the observed values and optimal parameters are returned.

    Parameters
    ----------
    opt_function : callable
        function to optimise
    x_data_orig : np.ndarray
        input data to the function - independent variables
    y_data : np.ndarray
        results that the optimisation should match - dependent variables
    x_init : np.ndarray
        initial guess for parameters
    low_bound : np.ndarray
        parameter low bound
    high_bound : np.ndarray
        parameter high bound
    opt_kwargs : dict
        optional meta-parameters to the optimisation function

    Returns
    -------
    tuple
        y_pred, y_res, opt_params : (np.ndarray, np.ndarray, np.ndarray).
        y_pred : predicted values,
        y_res : residual values,
        opt_params : optimal model parameters.
    """
    try:
        opt_params, _ = cast(
            tuple[npt.NDArray[np.float64], Any],
            curve_fit(
                opt_function,
                x_data_orig,
                y_data.flatten("F"),
                x_init,
                bounds=(low_bound, high_bound),
                method="trf",
                loss="soft_l1",
                **opt_kwargs,
            ),
        )

    except ValueError:
        raise ValueError(
            "gen_opt_routine: failed in optimisation step: {}".format(
                str(sys.exc_info())
            )
        )
    else:
        y_pred = np.reshape(
            opt_function(x_data_orig, *opt_params), y_data.shape, order="F"
        )
        y_res = y_pred - y_data

        # Alternative implementation, not shown to improve results
        # alt_opt_params = minimize(curve_fit_wrapper, x_init, args=(opt_function, x_data_orig, y_data.flatten('F')),
        #                          bounds=Bounds(low_bound, high_bound), method='SLSQP', options={'maxiter': 10000})
        # y_pred_1 = np.reshape(opt_function(x_data_orig, *alt_opt_params['x'], **opt_kwargs), y_data.shape, order='F')
        # y_res_1 = y_pred_1 - y_data
        # return y_pred_1, y_res_1, alt_opt_params['x']

        return y_pred, y_res, opt_params


def gen_mod_routine(
    opt_function: OptCallable,
    xdata_orig: npt.NDArray[np.float64],
    ydata_shape: tuple[int, int],
    opt_params: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Predict modelled values based on an earlier optimisation run for optimal model parameters.

    Parameters
    ----------
    opt_function : callable
        Function to optimise.
    xdata_orig : np.ndarray
        Input data to the function - independent variables.
    ydata_shape : (int, int)
        Shape of y_data.
    opt_params : np.ndarray
        Optimal model parameters.

    Returns
    -------
    np.ndarray
        Predicted values.
    """
    # Estimation of values
    return np.reshape(opt_function(xdata_orig, *opt_params), ydata_shape, order="F")


def gen_sub_routine(
    opt_function: OptCallable,
    xdata_orig: npt.NDArray[np.float64],
    xdata_new: npt.NDArray[np.float64],
    ydata: npt.NDArray[np.float64],
    opt_params: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """General substitution function based on a calibrated/optimised model and with two sets of input parameters.
    The substituted values are calculated as the original observations plus the difference of the two modelling
    steps.

    Parameters
    ----------
    opt_function : callable
        Function to optimise.
    xdata_orig : np.ndarray
        Input data to the function step 1 - independent variables.
    xdata_new : np.ndarray
        Input data to the function step 2 - independent variables.
    ydata : np.ndarray
        Original observed values step 1.
    opt_params : np.ndarray
        Set of optimal parameters to model.

    Returns
    -------
    tuple
        y_final, y_pred, y_res : (np.ndarray, np.ndarray, np.ndarray).
        Original observed data + difference in estimation between steps 0 and 1, y_pred - modelled data,
        y_res - residuals, opt_params - best parameter setting.
    """
    # Estimation of initial values
    y_pred = np.reshape(opt_function(xdata_orig, *opt_params), ydata.shape, order="F")
    # Estimation step for substituted fluid properties
    y_subst = np.reshape(opt_function(xdata_new, *opt_params), ydata.shape, "F")

    y_res = y_pred - ydata
    y_diff = y_subst - y_pred
    y_final = ydata + y_diff

    return y_final, y_pred, y_res


OptType = Literal["min", "exp", "pat_cem", "const_cem", "friable"]


class OptParamsDict(TypedDict, total=False):
    well_name: Required[str]
    opt_ver: Required[OptType]
    opt_vec: Required[npt.NDArray[np.float64]]
    f_ani: float
    f_con: float
    alpha_opt: npt.NDArray[np.float64]
    v_opt: float
    k_carb: float
    mu_carb: float
    rho_carb: float
    k_sh: float
    mu_sh: float
    rho_sh: float
    weight_k: float
    weight_mu: float
    shear_red: float
    frac_cem: float
    phi_c: float


class ParameterTranslationDict(TypedDict):
    opt_ver: str
    no_incl_sets: str
    ang_sym: str
    f_ani: str
    f_con: str
    alpha_opt: str
    v_opt: str
    k_carb: str
    mu_carb: str
    rho_carb: str
    k_sh: str
    mu_sh: str
    rho_sh: str
    k_sst: str
    mu_sst: str
    rho_sst: str
    frac_cem: str
    phi_c: str
    shear_red: str
    weight_k: str
    weight_mu: str


class ValueTranslationDict(TypedDict):
    ang_sym: float
    k_carb: float
    mu_carb: float
    rho_carb: float
    k_sh: float
    mu_sh: float
    rho_sh: float
    k_sst: float
    mu_sst: float
    rho_sst: float


class TypeTranslationDict(TypedDict):
    min: str
    exp: str
    pat_cem: str
    const_cem: str
    friable: str


def save_opt_params(
    opt_type: OptType,
    opt_params: npt.NDArray[np.float64],
    file_name: str = "opt_params.pkl",
    well_name: str = "Unknown well",
) -> None:
    """
    Utility to save optimal parameters as a pickle file in a more readable format so that the optimisation method can be
    recognised.

    Parameters
    ----------
    opt_type : str
        String defining optimisation type.
    opt_params : np.ndarray
        Numpy array with parameters from optimisation.
    file_name : str, optional
        File to save results to, by default 'opt_params.pkl'.
    well_name : str, optional
        Name of the well which is used in optimisation, by default 'Unknown well'.

    Raises
    ------
    ValueError
        If unknown optimisation opt_type.
    """
    # Save the optimal parameters with info
    if opt_type == "min":  # optimisation with mineral input from well
        opt_param_dict: OptParamsDict = {
            "well_name": well_name,
            "opt_ver": opt_type,
            "f_ani": opt_params[0],
            "f_con": opt_params[1],
            "alpha_opt": opt_params[2:4],
            "v_opt": opt_params[4],
            "opt_vec": opt_params,
        }
    elif opt_type == "exp":
        opt_param_dict = {
            "well_name": well_name,
            "opt_ver": opt_type,
            "f_ani": opt_params[0],
            "f_con": opt_params[1],
            "alpha_opt": opt_params[2:4],
            "v_opt": opt_params[4],
            "k_carb": opt_params[5],
            "mu_carb": opt_params[6],
            "rho_carb": opt_params[7],
            "k_sh": opt_params[8],
            "mu_sh": opt_params[9],
            "rho_sh": opt_params[10],
            "opt_vec": opt_params,
        }
    elif opt_type == "pat_cem":
        opt_param_dict = {
            "well_name": well_name,
            "opt_ver": opt_type,
            "weight_k": opt_params[0],
            "weight_mu": opt_params[1],
            "shear_red": opt_params[2],
            "frac_cem": opt_params[3],
            "opt_vec": opt_params,
        }
    elif opt_type == "const_cem":
        opt_param_dict = {
            "well_name": well_name,
            "opt_ver": opt_type,
            "phi_c": opt_params[0],
            "shear_red": opt_params[1],
            "frac_cem": opt_params[2],
            "opt_vec": opt_params,
        }
    elif opt_type == "friable":
        opt_param_dict = {
            "well_name": well_name,
            "opt_ver": opt_type,
            "phi_c": opt_params[0],
            "shear_red": opt_params[1],
            "opt_vec": opt_params,
        }
    else:
        raise ValueError(  # pyright: ignore[reportUnreachable] # kept for backward compatibility
            "save_opt_params: unknown optimisation opt_type: {}".format(opt_type)
        )

    with open(file_name, "wb") as file_out:
        pickle.dump(opt_param_dict, file_out)


def opt_param_info() -> tuple[
    ParameterTranslationDict, ValueTranslationDict, TypeTranslationDict
]:
    """Hard coded dictionaries returned.
    Returns
    -------
    tuple
        parameter_translation_dict, value_translation_dict, type_translation_dict.
    """
    parameter_translation_dict: ParameterTranslationDict = {
        "opt_ver": "Optimisation version",
        "no_incl_sets": "Number of inclusion sets",
        "ang_sym": "Angle of symmetry plane [°]",
        "f_ani": "Fraction of anisotropic inclusions",
        "f_con": "Fraction of connected inclusions",
        "alpha_opt": "Optimal aspect ratios for inclusion sets",
        "v_opt": "Ratio of volume for inclusion sets",
        "k_carb": "Matrix (carbonate) bulk modulus [Pa]",
        "mu_carb": "Matrix (carbonate) shear modulus [Pa]",
        "rho_carb": "Matrix (carbonate) density [kg/m^3]",
        "k_sh": "Mud/shale bulk modulus [Pa]",
        "mu_sh": "Mud/shale shear modulus [Pa]",
        "rho_sh": "Mud/shale density [kg/m^3]",
        "k_sst": "Sst bulk modulus [Pa]",
        "mu_sst": "Sst shear modulus [Pa]",
        "rho_sst": "Sst density [kg/m^3]",
        "frac_cem": "Cement fraction [fraction]",
        "phi_c": "Critical porosity [fraction]",
        "shear_red": "Reduction in tangential friction [fraction]",
        "weight_k": "Bulk modulus weight for constant cement model",
        "weight_mu": "Shear modulus weight for constant cement model",
    }
    value_translation_dict: ValueTranslationDict = {
        "ang_sym": 90.0,
        "k_carb": 95.0e9,
        "mu_carb": 45.0e9,
        "rho_carb": 2950.0,
        "k_sh": 35.0e9,
        "mu_sh": 20.0e9,
        "rho_sh": 2750.0,
        "k_sst": 45.0e9,
        "mu_sst": 50.0e9,
        "rho_sst": 2750.0,
    }
    type_translation_dict: TypeTranslationDict = {
        "min": "PETEC (Mineral input) optimisation",
        "exp": "Exploration type optimisation",
        "pat_cem": "Patchy cement model",
        "const_cem": "Constant cement model",
        "friable": "Friable sand model",
    }
    return parameter_translation_dict, value_translation_dict, type_translation_dict


def load_opt_params(
    file_name: str | Path,
) -> tuple[
    OptType,
    npt.NDArray[np.float64],
    OptParamsDict,
]:
    """Utility to load parameter file from optimisation run.

    Parameters
    ----------
    file_name : str
        Input file name including path.

    Returns
    -------
    tuple
        opt_type: model type, no_sets: number of inclusion sets, opt_param: with all parameters for model.
    """
    with open(file_name, "rb") as fin:
        param_dict: OptParamsDict = pickle.load(fin)
        opt_type = param_dict["opt_ver"]
        opt_param = param_dict["opt_vec"]
        opt_dict = param_dict

        return opt_type, opt_param, opt_dict


def opt_param_to_ascii(
    in_file: str | Path,
    display_results: bool = True,
    out_file: str | Path | None = None,
    well_name: str = "Unknown well",
    **kwargs: Any,
) -> None:
    """Utility to convert stored optimised parameters to ascii and display results or save to file.

    Parameters
    ----------
    in_file : str
        File name for stored optimised parameters.
    display_results : bool
        Display results on screen, default True.
    out_file : str or None
        Optional store optimised parameters in ascii file.
    well_name : str
        Optional name of the well that is used in optimisation.
    """
    with open(in_file, "rb") as f_in:
        param_dict = pickle.load(f_in)
        if well_name.lower() == "unknown well":
            well_name = param_dict.pop("well_name", "Unknown Well")

        (
            parameter_translation_dict,
            value_translation_dict,
            type_translation_dict,
        ) = opt_param_info()

        item: list[str] = []
        value: list[str] = []
        disp_string = ""
        for opt_key, opt_value in param_dict.items():
            if opt_key in parameter_translation_dict:
                if opt_key in value_translation_dict:
                    opt_value = opt_value * value_translation_dict[opt_key]
                    opt_str = f" {opt_value:.4f}"
                elif opt_key == "opt_ver":
                    opt_str = type_translation_dict[opt_value]
                elif opt_key == "v_opt":
                    opt_value = np.append(opt_value, 1.0 - np.sum(opt_value))
                    opt_str = f" {opt_value:}"
                else:
                    if isinstance(opt_value, float):
                        opt_str = f" {opt_value:.4f}"
                    else:
                        opt_str = f" {opt_value:}"
                item.append(f"{parameter_translation_dict[opt_key]}: ")
                value.append(opt_str)
                disp_string += f"{parameter_translation_dict[opt_key]}: {opt_str}\n"
        info_array = np.stack((item, value), axis=1)

        if display_results:
            from tkinter import END, Entry, Tk

            class Table:
                def __init__(
                    self,
                    tk_root: Tk,
                    no_rows: int,
                    no_cols: int,
                    info: npt.NDArray[Any],
                ):
                    # code for creating table
                    str_len = np.vectorize(len)
                    text_justify: list[Literal["right", "left"]] = ["right", "left"]
                    for i in range(no_rows):
                        for j in range(no_cols):
                            just = text_justify[0] if j == 0 else text_justify[1]
                            max_len = np.max(str_len(info[:, j]))
                            self.e: Entry = Entry(
                                root,
                                width=max_len + 2,
                                fg="black",
                                font=("Consolas", 11, "normal"),
                                justify=just,
                            )
                            self.e.grid(row=i, column=j)
                            self.e.insert(END, info[i][j])

            root = Tk(**kwargs)
            if well_name.lower() == "unknown well":
                root.title("T Matrix Optimised Parameters")
            else:
                root.title(well_name)
            if sys.platform.startswith("win"):
                ico_file = os.path.join(os.path.dirname(__file__), "Equinor_logo.ico")
                root.iconbitmap(ico_file)
            _ = Table(root, info_array.shape[0], info_array.shape[1], info_array)
            _ = root.attributes("-topmost", True)
            root.mainloop()

        if out_file is not None:
            with open(out_file, "w") as f_out:
                _ = f_out.write(disp_string)

        return



import inspect
import math
import re
from typing import Any
from warnings import warn

import numpy as np
import numpy.typing as npt
import pandas as pd

from rock_physics_open.equinor_utilities.various_utilities import disp_result_stats

from .snapshots import get_snapshot_name


def compare_snapshots(
    test_results: tuple[npt.NDArray[np.float64] | pd.DataFrame, ...]
    | npt.NDArray[np.float64],
    saved_results: tuple[npt.NDArray[np.float64] | pd.DataFrame, ...],
    name_arr: list[str] | None = None,
    display_results: bool = False,
) -> bool:
    test_results = _validate_input(test_results, saved_results)

    if display_results:
        title = str(inspect.stack()[1].function)
        if not name_arr:
            name_arr = [f"arr_{i}" for i in range(len(test_results))]
        disp_result_stats(title, test_results, name_arr)

    r_tol = 0.01
    equal_nan = True
    no_difference_found = True

    for i, (test_item, saved_item) in enumerate(zip(test_results, saved_results)):
        try:
            if name_arr:
                err_msg = f"saved and generated result for {name_arr[i]} differ"
            else:
                err_msg = f"saved result for variable {i} differ"
            np.testing.assert_allclose(
                test_item,
                saved_item,
                rtol=r_tol,
                equal_nan=equal_nan,
                err_msg=err_msg,
            )
        except AssertionError as error:
            open_mode = "w" if no_difference_found else "a"
            no_difference_found = False
            warn(f"comparison test failed for item {i}: {error}")
            log_file = re.sub("npz", "log", get_snapshot_name(step=2))

            with open(log_file, open_mode) as file:
                _ = file.write(
                    f"Test function: {get_snapshot_name(include_extension=False, include_snapshot_dir=False, include_filename=False)} \n"
                )
                if name_arr:
                    _ = file.write(f"Test variable: {name_arr[i]} \n")
                else:
                    _ = file.write(f"Test variable number: {i} \n")

                for line in str(error).splitlines():
                    mismatched_elements_index = (
                        line.replace(" ", "").lower().find("mismatchedelements")
                    )
                    if mismatched_elements_index != -1:
                        _ = file.write(line + "\n")
                        continue

                    max_abs_diff_index = (
                        line.replace(" ", "").lower().find("maxabsolutedifference")
                    )
                    if max_abs_diff_index != -1:
                        _ = file.write(line + "\n")
                        continue

                    max_rel_diff_index = (
                        line.replace(" ", "").lower().find("maxrelativedifference")
                    )
                    if max_rel_diff_index != -1:
                        _ = file.write(line + "\n")
                        continue

                    reg_index = re.search(r"differ", line)

                    if reg_index:
                        if isinstance(test_item, np.ndarray) and isinstance(
                            saved_item, np.ndarray
                        ):
                            differences, num_nans = _compare_ndarray(
                                saved_array=saved_item,
                                test_array=test_item,
                                eq_nan=equal_nan,
                                rel_tol=r_tol,
                            )
                        elif isinstance(test_item, pd.DataFrame) and isinstance(
                            saved_item, pd.DataFrame
                        ):
                            differences, num_nans = _compare_df(
                                saved_results=saved_item,
                                test_results=test_item,
                                equal_nan=equal_nan,
                                r_tol=r_tol,
                            )
                        else:
                            raise TypeError(
                                f"Unsupported data type for comparison: {type(test_item)}"
                            )
                        _ = file.write(
                            "Number of NaN elements: " + str(num_nans) + "\n"
                        )
                        _ = file.write("Index:\t\tSaved:\t\tGenerated:\n")

                        # Write test results and saved results differences to file
                        if len(differences) > 0:
                            tab = "\t"
                            for difference in differences:
                                _ = file.write(
                                    f"{tab}[{difference[0]:4}]=> {difference[1]:.8g} != {difference[2]:.8g}\n"
                                )
                            _ = file.write(f"{'_' * 40}\n")
    return no_difference_found


def _compare_ndarray(
    saved_array: npt.NDArray[np.float64],
    test_array: npt.NDArray[np.float64],
    eq_nan: bool,
    rel_tol: float,
) -> tuple[list[list[Any]], int]:
    differ_indexes = np.where(saved_array != test_array)[0]
    differences: list[list[Any]] = []
    num_nans = 0

    for index in differ_indexes:
        if eq_nan and (
            np.isnan(test_array[int(index)]) and np.isnan(saved_array[int(index)])
        ):
            num_nans += 1
        elif math.isclose(
            saved_array[int(index)], test_array[int(index)], rel_tol=rel_tol
        ):
            pass
        else:
            differences.append([index, saved_array[index], test_array[index]])
    return differences, num_nans


def _compare_df(
    saved_results: pd.DataFrame,
    test_results: pd.DataFrame,
    equal_nan: bool,
    r_tol: float,
) -> tuple[list[list[Any]], int]:
    return _compare_ndarray(
        saved_array=saved_results.to_numpy().flatten(),
        test_array=test_results.to_numpy().flatten(),
        eq_nan=equal_nan,
        rel_tol=r_tol,
    )


def _validate_input(
    test_obj: tuple[npt.NDArray[np.float64] | pd.DataFrame, ...]
    | npt.NDArray[np.float64]
    | pd.DataFrame,
    saved_obj: tuple[npt.NDArray[np.float64] | pd.DataFrame, ...],
) -> tuple[npt.NDArray[np.float64] | pd.DataFrame, ...]:
    # Check for compatibility of test results and stored data
    if isinstance(test_obj, (np.ndarray, pd.DataFrame)):
        return_test_obj = (test_obj,)
    else:
        return_test_obj = test_obj
    if isinstance(return_test_obj, (tuple, list)):  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
        if len(saved_obj) != len(return_test_obj):
            raise ValueError(
                f"unable to compare snapshots, different number of saved: ({len(saved_obj)}) and generated results ({len(test_obj)})"
            )
    else:
        raise ValueError(
            f"test_obj should be one of list, tuple, numpy array or pandas DataFrame, is {type(test_obj)}"
        )
    return return_test_obj
